# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json
import os
import re
import subprocess
import time
from tempfile import NamedTemporaryFile

from kafka.tools import log
from kafka.tools.exceptions import ReassignmentFailedException
from kafka.tools.models import BaseModel


class Reassignment(BaseModel):
    equality_attrs = ['partitions']

    def __init__(self, partitions, pause_time=10):
        self.partitions = partitions
        self.pause_time = pause_time
        self.status_re = re.compile('.*Reassignment of partition.*?\s+(failed|still in progress|completed successfully)')

    def __repr__(self):
        return json.dumps(self.dict_for_reassignment())

    def dict_for_reassignment(self):
        reassignment = {'partitions': [], 'version': 1}
        for partition in self.partitions:
            reassignment['partitions'].append(partition.dict_for_reassignment())
        return reassignment

    def execute(self, num, total, zookeeper, tools_path, plugins=[], dry_run=True):
        for plugin in plugins:
            plugin.before_execute_batch(num)
        if not dry_run:
            self._execute(num, total, zookeeper, tools_path)
        for plugin in plugins:
            plugin.after_execute_batch(num)

    def _execute(self, num, total, zookeeper, tools_path):
        with NamedTemporaryFile(mode='w') as assignfile:
            json.dump(self.dict_for_reassignment(), assignfile)
            assignfile.flush()
            FNULL = open(os.devnull, 'w')
            proc = subprocess.Popen(['{0}/kafka-reassign-partitions.sh'.format(tools_path), '--execute',
                                     '--zookeeper', zookeeper,
                                     '--reassignment-json-file', assignfile.name],
                                    stdout=FNULL, stderr=FNULL)
            proc.wait()

            # Wait until finished
            while True:
                remaining_partitions = self.check_completion(zookeeper, tools_path, assignfile.name)
                if remaining_partitions == 0:
                    break

                log.info('Partition reassignment {0}/{1} in progress [ {2}/{3} partitions remain ]. Sleeping {4} seconds'.format(num,
                                                                                                                                 total,
                                                                                                                                 remaining_partitions,
                                                                                                                                 len(self.partitions),
                                                                                                                                 self.pause_time))
                time.sleep(self.pause_time)

    def process_verify_match(self, line):
        match_obj = self.status_re.match(line)
        if match_obj:
            if match_obj.group(1) == 'failed':
                return -1
            elif match_obj.group(1) == 'still in progress':
                return 1
        return 0

    def check_completion(self, zookeeper, tools_path, assign_filename):
        FNULL = open(os.devnull, 'w')
        proc = subprocess.Popen(['{0}/kafka-reassign-partitions.sh'.format(tools_path), '--verify',
                                 '--zookeeper', zookeeper,
                                 '--reassignment-json-file', assign_filename],
                                stdout=subprocess.PIPE, stderr=FNULL)
        lines = proc.stdout.readlines()

        remaining_count = 0
        for line in lines:
            count = self.process_verify_match(line.decode())
            if count < 0:
                raise ReassignmentFailedException("The reassignment in progress failed with the following verification output:\n{0}".format(lines))
            remaining_count += count

        return remaining_count



import time
import unittest
from mock import MagicMock

from tests.tools.client.fixtures import describe_groups, describe_groups_error

from kafka.tools.client import Client
from kafka.tools.exceptions import GroupError
from kafka.tools.models.broker import Broker
from kafka.tools.models.group import Group
from kafka.tools.protocol.requests.describe_groups_v0 import DescribeGroupsV0Request


class InterfaceGroupsTests(unittest.TestCase):
    def setUp(self):
        # Dummy client for testing - we're not going to connect that bootstrap broker
        self.client = Client()
        self.client._connected = True

        # Two brokers for the client
        broker = Broker('host1.example.com', id=1, port=8031)
        broker.rack = 'rack1'
        self.client.cluster.add_broker(broker)
        broker = Broker('host2.example.com', id=101, port=8032)
        broker.rack = 'rack1'
        self.client.cluster.add_broker(broker)

        self.describe_groups = describe_groups()
        self.describe_groups_error = describe_groups_error()

    def test_list_groups(self):
        self.client.cluster.add_group(Group('group2'))

        self.client._maybe_update_groups_list = MagicMock()
        self.client._maybe_update_groups_list.return_value = 312

        groups, errs = self.client.list_groups()
        self.client._maybe_update_groups_list.assert_called_once_with(True)
        assert errs == 312
        assert groups == ['group2']

    def test_list_groups_cache(self):
        self.client._maybe_update_groups_list = MagicMock()

        self.client.list_groups(cache=False)
        self.client._maybe_update_groups_list.assert_called_once_with(False)

    def test_get_group(self):
        def add_group_to_cluster(group_name, request):
            # _send_group_aware_request has a side effect of creating the group and setting the coordinator
            self.client.cluster.add_group(Group(group_name))
            self.client.cluster.groups[group_name].coordinator = self.client.cluster.brokers[1]
            return self.describe_groups

        self.client._send_group_aware_request = MagicMock()
        self.client._send_group_aware_request.side_effect = add_group_to_cluster

        val = self.client.get_group('testgroup')

        self.client._send_group_aware_request.assert_called_once()
        assert self.client._send_group_aware_request.call_args[0][0] == 'testgroup'
        assert isinstance(self.client._send_group_aware_request.call_args[0][1], DescribeGroupsV0Request)

        assert isinstance(val, Group)
        assert val.name == 'testgroup'
        print(val.coordinator)
        print(self.client.cluster.brokers[1])
        assert val.coordinator == self.client.cluster.brokers[1]

    def test_get_group_existing_cached(self):
        group = Group('testgroup')
        group.coordinator = self.client.cluster.brokers[1]
        group._last_updated = time.time()
        self.client.cluster.add_group(group)

        self.client._send_group_aware_request = MagicMock()
        self.client._send_group_aware_request.return_value = self.describe_groups
        self.client._update_groups_from_describe = MagicMock()
        val = self.client.get_group('testgroup')

        self.client._send_group_aware_request.assert_not_called()
        self.client._update_groups_from_describe.assert_not_called()

        assert isinstance(val, Group)
        assert val.name == 'testgroup'
        assert val.coordinator == self.client.cluster.brokers[1]

    def test_get_group_error(self):
        group = Group('badgroup')
        group.coordinator = self.client.cluster.brokers[1]
        self.client.cluster.add_group(group)

        self.client._send_group_aware_request = MagicMock()
        self.client._send_group_aware_request.return_value = self.describe_groups_error

        self.assertRaises(GroupError, self.client.get_group, 'badgroup', cache=False)
        self.client._send_group_aware_request.assert_called_once()

    def test_get_group_nonexistent(self):
        self.client._send_group_aware_request = MagicMock()
        self.client._send_group_aware_request.side_effect = GroupError

        self.assertRaises(GroupError, self.client.get_group, 'nonexistentgroup')
        self.client._send_group_aware_request.assert_called_once()



import argparse
import inspect
import sys
import unittest

from contextlib import contextmanager

import kafka.tools.assigner.actions

from kafka.tools.assigner.arguments import set_up_arguments, CSVAction
from kafka.tools.modules import get_modules
from kafka.tools.assigner.plugins import PluginModule


@contextmanager
def redirect_err_output():
    current_err = sys.stderr
    try:
        sys.stderr = sys.stdout
        yield
    finally:
        sys.stderr = current_err


class ArgumentTests(unittest.TestCase):
    def setUp(self):
        self.null_plugin = PluginModule()

    def create_action_map(self):
        self.action_map = dict((cls.name, cls) for cls in get_modules(kafka.tools.assigner.actions, kafka.tools.assigner.actions.ActionModule))

    def test_get_arguments_none(self):
        sys.argv = ['kafka-assigner']
        with redirect_err_output():
            self.assertRaises(SystemExit, set_up_arguments, {}, {}, [self.null_plugin])

    def test_get_modules(self):
        self.create_action_map()
        assert 'elect' in self.action_map
        assert inspect.isclass(self.action_map['elect'])

    def test_get_arguments_minimum(self):
        self.create_action_map()
        sys.argv = ['kafka-assigner', '--zookeeper', 'zkhost1.example.com:2181', 'elect']
        args = set_up_arguments(self.action_map, {}, [self.null_plugin])
        assert args.action == 'elect'
        assert args.zookeeper == 'zkhost1.example.com:2181'
        assert args.leadership is False
        assert args.generate is False
        assert args.execute is False
        assert args.property == []
        assert args.size is False
        assert args.skip_ple is False
        assert args.sizer == 'ssh'
        assert args.moves == 10
        assert args.ple_size == 2000
        assert args.ple_wait == 120

    def test_csv_action_single(self):
        aparser = argparse.ArgumentParser()
        aparser.add_argument('--foo', action=CSVAction)
        args = aparser.parse_args(['--foo', '1'])
        assert args.foo == ['1']

    def test_csv_action_nargs(self):
        aparser = argparse.ArgumentParser()
        self.assertRaises(ValueError, aparser.add_argument, '--foo', action=CSVAction, nargs='*')

    def test_csv_action_two_args(self):
        aparser = argparse.ArgumentParser()
        aparser.add_argument('--foo', action=CSVAction)
        args = aparser.parse_args(['--foo', '1', '--foo', '2'])
        assert args.foo == ['1', '2']

    def test_csv_action_commas(self):
        aparser = argparse.ArgumentParser()
        aparser.add_argument('--foo', action=CSVAction)
        args = aparser.parse_args(['--foo', '1,2'])
        assert args.foo == ['1', '2']



import sys
import unittest

from argparse import Namespace
from .fixtures import set_up_cluster, set_up_subparser

from kafka.tools.assigner.actions.elect import ActionElect


class ActionElectTests(unittest.TestCase):
    def setUp(self):
        self.cluster = set_up_cluster()
        (self.parser, self.subparsers) = set_up_subparser()
        self.args = Namespace(exclude_topics=[])

    def test_create_class(self):
        action = ActionElect(self.args, self.cluster)
        assert isinstance(action, ActionElect)

    def test_configure_args(self):
        ActionElect.configure_args(self.subparsers)
        sys.argv = ['kafka-assigner', 'elect']
        parsed_args = self.parser.parse_args()
        assert parsed_args.action == 'elect'

    def test_process_cluster(self):
        action = ActionElect(self.args, self.cluster)
        action.process_cluster()
        assert True



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import time

from kafka.tools.exceptions import OffsetError
from kafka.tools.models import BaseModel
from kafka.tools.models.partition import Partition
from kafka.tools.utilities import raise_if_error


class Topic(BaseModel):
    equality_attrs = ['name']

    def __init__(self, name, partitions):
        self.name = name
        self.partitions = []
        self.cluster = None
        self.internal = False
        self.retention = 1
        self._last_updated = time.time()

        for i in range(partitions):
            self.add_partition(Partition(self, i))

    # Shallow copy - do not copy partitions (zero partitions)
    def copy(self):
        newtopic = Topic(self.name, 0)
        newtopic.cluster = self.cluster
        return newtopic

    def add_partition(self, partition):
        self.partitions.append(partition)

    def updated_since(self, check_time):
        return check_time <= self._last_updated

    def assure_has_partitions(self, target_count):
        """
        Assure that the topic has only the specified number of partitions, adding or deleting as needed

        Args:
            target_count (int): the number of partitions to have
        """
        while len(self.partitions) < target_count:
            partition = Partition(self.name, len(self.partitions))
            self.add_partition(partition)

        # While Kafka doesn't support partition deletion (only topics), it's possible for a topic
        # to be deleted and recreated with a smaller partition count before we see that it's been
        # deleted. This would look like partition deletion, so we should support it.
        while len(self.partitions) > target_count:
            partition = self.partitions.pop()
            partition.delete_replicas(0)


class TopicOffsets:
    def __init__(self, topic):
        self.topic = topic
        self.partitions = [-1 for i in range(len(topic.partitions))]

    def set_offsets_from_list(self, partitions):
        """
        Given a partition_responses object from a ListOffsets response, update the offsets
        with the values in the response

        Args:
            partitions (Array): the partition_response object from a ListOffsets response

        Raises:
            OffsetError: If there was a failure retrieving any of the offsets
        """
        for partition in partitions:
            raise_if_error(OffsetError, partition['error'])
            self.partitions[partition['partition'].value()] = partition['offsets'][0].value()

    def set_offsets_from_fetch(self, partitions):
        """
        Given a partition_responses object from a OffsetFetch response, update the offsets
        with the values in the response

        Args:
            partitions (Array): the partition_response object from a OffsetFetch response

        Raises:
            OffsetError: If there was a failure retrieving any of the offsets
        """
        for partition in partitions:
            raise_if_error(OffsetError, partition['error'])
            self.partitions[partition['partition'].value()] = partition['offset'].value()



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.delete_topics_v0 import DeleteTopicsV0Request


class DeleteTopicsV0RequestTests(unittest.TestCase):
    def test_process_arguments(self):
        val = DeleteTopicsV0Request.process_arguments(['3296', 'topicname', 'anothertopic'])
        assert val == {'timeout': 3296, 'topics': ['topicname', 'anothertopic']}

    def test_process_arguments_notimeout(self):
        self.assertRaises(ArgumentError, DeleteTopicsV0Request.process_arguments, ['topicname', 'anothertopic'])

    def test_process_arguments_notopic(self):
        self.assertRaises(ArgumentError, DeleteTopicsV0Request.process_arguments, ['3296'])



import unittest
from mock import MagicMock

from tests.tools.client.fixtures import topic_metadata, offset_fetch

from kafka.tools.client import Client
from kafka.tools.exceptions import GroupError
from kafka.tools.models.group import Group
from kafka.tools.models.topic import TopicOffsets
from kafka.tools.protocol.requests.offset_fetch_v1 import OffsetFetchV1Request


class InterfaceOffsetsTests(unittest.TestCase):
    def setUp(self):
        # Dummy client for testing - we're not going to connect that bootstrap broker
        self.client = Client()
        self.client._connected = True

        # Get the broker and topic from a metadata update
        self.client._update_from_metadata(topic_metadata())

        self.group = Group('testgroup')
        self.offset_fetch = offset_fetch()

    def test_get_offsets_for_topics(self):
        self.client._maybe_update_metadata_for_topics = MagicMock()
        self.client._send_list_offsets_to_brokers = MagicMock()
        self.client._send_list_offsets_to_brokers.return_value = {'topic1': 'responseobj'}

        val = self.client.get_offsets_for_topics(['topic1'])
        self.client._maybe_update_metadata_for_topics.assert_called_once_with(['topic1'])
        self.client._send_list_offsets_to_brokers.assert_called_once()

        values = self.client._send_list_offsets_to_brokers.call_args[0][0]
        assert len(values) == 2
        for broker_id in (1, 101):
            assert values[broker_id]['replica_id'] == -1
            assert len(values[broker_id]['topics']) == 1
            assert values[broker_id]['topics'][0]['topic'] == 'topic1'
            assert len(values[broker_id]['topics'][0]['partitions']) == 1
            assert values[broker_id]['topics'][0]['partitions'][0]['timestamp'] == Client.OFFSET_LATEST
            assert values[broker_id]['topics'][0]['partitions'][0]['max_num_offsets'] == 1
        assert values[1]['topics'][0]['partitions'][0]['partition'] == 0
        assert values[101]['topics'][0]['partitions'][0]['partition'] == 1

        assert val == {'topic1': 'responseobj'}

    def test_get_offsets_for_topic(self):
        self.client.get_offsets_for_topics = MagicMock()
        self.client.get_offsets_for_topics.return_value = {'topic1': 'thingitreturns'}
        val = self.client.get_offsets_for_topic('topic1')
        self.client.get_offsets_for_topics.assert_called_once_with(['topic1'], Client.OFFSET_LATEST)
        assert val == 'thingitreturns'

    def test_get_offsets_for_topic_bad_timestamp(self):
        self.assertRaises(TypeError, self.client.get_offsets_for_topics, ['topic1'], timestamp='notanint')

    def test_get_topics_for_group_string(self):
        val = self.client._get_topics_for_group(self.group, 'testtopic')
        assert val == ['testtopic']

    def test_get_topics_for_group_list(self):
        val = self.client._get_topics_for_group(self.group, ['topica', 'topicb'])
        assert val == ['topica', 'topicb']

    def test_get_topics_for_group_empty(self):
        self.assertRaises(GroupError, self.client._get_topics_for_group, self.group, [])

    def test_get_topics_for_group_subscribed(self):
        self.group.subscribed_topics = MagicMock()
        self.group.subscribed_topics.return_value = ['topica']
        val = self.client._get_topics_for_group(self.group, None)

        self.group.subscribed_topics.assert_called_once_with()
        assert val == ['topica']

    def test_get_offsets_for_group(self):
        self.client.get_group = MagicMock()
        self.client.get_group.return_value = self.group
        self.client._get_topics_for_group = MagicMock()
        self.client._get_topics_for_group.return_value = ['topic1']
        self.client._maybe_update_metadata_for_topics = MagicMock()
        self.client._send_group_aware_request = MagicMock()
        self.client._send_group_aware_request.return_value = self.offset_fetch

        val = self.client.get_offsets_for_group('testgroup')

        self.client.get_group.assert_called_once_with('testgroup')
        self.client._get_topics_for_group.assert_called_once_with(self.group, None)
        self.client._maybe_update_metadata_for_topics.assert_called_once_with(['topic1'])

        self.client._send_group_aware_request.assert_called_once()
        assert self.client._send_group_aware_request.call_args[0][0] == 'testgroup'
        req = self.client._send_group_aware_request.call_args[0][1]
        assert isinstance(req, OffsetFetchV1Request)
        assert req['group_id'].value() == 'testgroup'
        assert len(req['topics']) == 1
        assert req['topics'][0]['topic'].value() == 'topic1'
        assert len(req['topics'][0]['partitions']) == 2
        assert req['topics'][0]['partitions'][0].value() == 0
        assert req['topics'][0]['partitions'][1].value() == 1

        assert 'topic1' in val
        assert isinstance(val['topic1'], TopicOffsets)
        assert val['topic1'].partitions == [4829, 8904]

    def test_set_offsets_for_group_bad_offsets(self):
        self.assertRaises(TypeError, self.client.set_offsets_for_group, 'testgroup', 'notalist')

    def test_set_offsets_for_group_offsets_none(self):
        self.assertRaises(TypeError, self.client.set_offsets_for_group, 'testgroup', None)

    def test_set_offsets_for_group_not_empty(self):
        self.group.state = 'Stable'

        self.client.get_group = MagicMock()
        self.client.get_group.return_value = self.group

        self.assertRaises(GroupError, self.client.set_offsets_for_group, 'testgroup', [])
        self.client.get_group.assert_called_once_with('testgroup')

    def test_set_offsets_for_group(self):
        self.client.get_group = MagicMock()
        self.client.get_group.return_value = self.group
        self.client._send_set_offset_request = MagicMock()
        self.client._send_set_offset_request.return_value = 'sendresponse'
        self.client._parse_set_offset_response = MagicMock()
        self.client._parse_set_offset_response.return_value = {'topic1': [0, 0]}

        offsets = TopicOffsets(self.client.cluster.topics['topic1'])
        offsets.partitions[0] = 2342
        offsets.partitions[1] = 8793
        val = self.client.set_offsets_for_group('testgroup', [offsets])

        assert val == {'topic1': [0, 0]}
        self.client.get_group.assert_called_once_with('testgroup')
        self.client._send_set_offset_request.assert_called_once_with('testgroup', [offsets])
        self.client._parse_set_offset_response.assert_called_once_with('sendresponse')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from collections import deque
from kafka.tools.assigner.actions import ActionModule
from kafka.tools.exceptions import ConfigurationException, NotEnoughReplicasException


class ActionRemove(ActionModule):
    name = "remove"
    helpstr = "Move partitions from one broker to one or more other brokers (maintaining RF)"

    def __init__(self, args, cluster):
        super(ActionRemove, self).__init__(args, cluster)

        self.check_brokers(type_str="Brokers to remove")
        if len(set(self.args.to_brokers) & set(self.args.brokers)) != 0:
            raise ConfigurationException("Brokers to remove were specified in the target broker list as well")

        self.brokers = args.brokers
        self.to_brokers = args.to_brokers
        if len(self.to_brokers) == 0:
            self.to_brokers = list(set(self.cluster.brokers.keys()) - set(self.brokers))
        if set(self.args.to_brokers) & set(self.cluster.brokers.keys()) != set(self.args.to_brokers):
            raise ConfigurationException("Target broker are not in the brokers list for this cluster")

    @classmethod
    def _add_args(cls, parser):
        parser.add_argument('-b', '--brokers', help="List of Broker IDs to remove", required=True, type=int, nargs='*')
        parser.add_argument('-t', '--to_brokers', help="List of Broker IDs to move partitions to (defaults to whole cluster)",
                            required=False, type=int, nargs='*', default=[])

    def process_cluster(self):
        # Make a deque for the target brokers so we can round-robin assignments
        todeque = deque(self.to_brokers)

        for broker_id in self.brokers:
            broker = self.cluster.brokers[broker_id]
            for position in broker.partitions:
                iterlist = list(broker.partitions[position])
                for partition in iterlist:
                    if partition.topic.name in self.args.exclude_topics:
                        continue

                    # Find a new replica for this partition
                    newreplica = None
                    attempts = 0
                    while attempts < len(todeque):
                        proposed = todeque.popleft()
                        todeque.append(proposed)
                        proposed_broker = self.cluster.brokers[proposed]
                        if proposed_broker not in partition.replicas:
                            newreplica = proposed_broker
                            break
                        attempts += 1

                    if newreplica is None:
                        raise NotEnoughReplicasException("Cannot find a new broker for {0}:{1} with replica list {2}".format(partition.topic.name,
                                                                                                                             partition.num,
                                                                                                                             partition.replicas))

                    # Replace the broker coming out with the new one
                    partition.swap_replicas(broker, newreplica)



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.list_offset_v0 import _parse_next_topic, ListOffsetV0Request


class ListOffsetV0RequestTest(unittest.TestCase):
    def test_parse_next_topic(self):
        val, rest = _parse_next_topic(['topicname', '4,2,8'])
        assert val == {'topic': 'topicname', 'partitions': [{'partition': 4, 'timestamp': 2, 'max_num_offsets': 8}]}
        assert rest == []

    def test_parse_next_topic_multiple(self):
        val, rest = _parse_next_topic(['topicname', '4,2,8', '9,3,2'])
        assert val == {'topic': 'topicname', 'partitions': [{'partition': 4, 'timestamp': 2, 'max_num_offsets': 8},
                       {'partition': 9, 'timestamp': 3, 'max_num_offsets': 2}]}
        assert rest == []

    def test_parse_next_topic_remainder(self):
        val, rest = _parse_next_topic(['topicname', '4,2,8', 'nexttopic', '9,3,2'])
        assert val == {'topic': 'topicname', 'partitions': [{'partition': 4, 'timestamp': 2, 'max_num_offsets': 8}]}
        assert rest == ['nexttopic', '9,3,2']

    def test_parse_next_topic_nopartitions(self):
        self.assertRaises(ArgumentError, _parse_next_topic, ['topicname'])

    def test_parse_next_topic_short_partitions(self):
        self.assertRaises(ArgumentError, _parse_next_topic, ['topicname', '4,2'])

    def test_parse_next_topic_extra_partitions(self):
        self.assertRaises(ArgumentError, _parse_next_topic, ['topicname', '4,2,8,9'])

    def test_parse_next_topic_nonnumeric(self):
        self.assertRaises(ArgumentError, _parse_next_topic, ['topicname', 'foo,2,8'])
        self.assertRaises(ArgumentError, _parse_next_topic, ['topicname', '4,foo,8'])
        self.assertRaises(ArgumentError, _parse_next_topic, ['topicname', '4,2,foo'])

    def test_process_arguments(self):
        val = ListOffsetV0Request.process_arguments(['-1', 'topicname', '4,2,8', 'nexttopic', '9,3,2'])
        assert val == {'replica_id': -1,
                       'topics': [{'topic': 'topicname', 'partitions': [{'partition': 4, 'timestamp': 2, 'max_num_offsets': 8}]},
                                  {'topic': 'nexttopic', 'partitions': [{'partition': 9, 'timestamp': 3, 'max_num_offsets': 2}]}]}

    def test_process_arguments_nonnumeric(self):
        self.assertRaises(ArgumentError, ListOffsetV0Request.process_arguments, ['foo', 'topicname', '4,2,8'])

    def test_process_arguments_notenough(self):
        self.assertRaises(ArgumentError, ListOffsetV0Request.process_arguments, ['-1', 'topicname'])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kazoo.client import KazooClient
from kazoo.exceptions import KazooException

from kafka.tools import log
from kafka.tools.exceptions import ZookeeperException, ClusterConsistencyException
from kafka.tools.models import BaseModel
from kafka.tools.models.broker import Broker
from kafka.tools.models.topic import Topic
from kafka.tools.utilities import json_loads


def add_brokers_from_zk(cluster, zk):
    for b in zk.get_children("/brokers/ids"):
        broker_data, bstat = zk.get("/brokers/ids/{0}".format(b))
        cluster.add_broker(Broker.create_from_json(int(b), broker_data))
    if cluster.num_brokers() == 0:
        raise ZookeeperException("The cluster specified does not have any brokers")


def add_topic_with_replicas(cluster, topic, topic_data):
    newtopic = Topic(topic, len(topic_data['partitions']))
    for partition in topic_data['partitions']:
        for i, replica in enumerate(topic_data['partitions'][partition]):
            if replica not in cluster.brokers:
                # Hit a replica that's not in the ID list (which means it's dead)
                # We'll add it, but trying to get sizes will fail as we don't have a hostname
                cluster.add_broker(Broker(None, id=replica))
            newtopic.partitions[int(partition)].add_replica(cluster.brokers[replica], i)
    cluster.add_topic(newtopic)


def set_topic_retention(topic, zk):
    try:
        zdata, zstat = zk.get("/config/topics/{0}".format(topic.name))
        tdata = json_loads(zdata)
        topic.retention = int(tdata['config']['retention.ms'])
    except (KeyError, ValueError, KazooException):
        # If we can't get the config override for any reason, just stick with whatever the default is
        pass


class Cluster(BaseModel):
    equality_attrs = ['brokers', 'topics']

    def __init__(self, retention=1):
        self.brokers = {}
        self.topics = {}
        self.groups = {}
        self.retention = retention

    @classmethod
    def create_from_zookeeper(cls, zkconnect, default_retention=1, fetch_topics=True):
        log.info("Connecting to zookeeper {0}".format(zkconnect))
        try:
            zk = KazooClient(zkconnect)
            zk.start()
        except Exception as e:
            raise ZookeeperException("Cannot connect to Zookeeper: {0}".format(e))

        # Get broker list
        cluster = cls(retention=default_retention)
        add_brokers_from_zk(cluster, zk)

        # Get current partition state
        if fetch_topics:
            log.info("Getting partition list from Zookeeper")
            for topic in zk.get_children("/brokers/topics"):
                zdata, zstat = zk.get("/brokers/topics/{0}".format(topic))
                add_topic_with_replicas(cluster, topic, json_loads(zdata))
                set_topic_retention(cluster.topics[topic], zk)

            if cluster.num_topics() == 0:
                raise ZookeeperException("The cluster specified does not have any topics")

        log.info("Closing connection to zookeeper")
        zk.stop()
        zk.close()

        return cluster

    def clone(self):
        newcluster = Cluster()
        newcluster.retention = self.retention

        # We're not going to clone in the subclasses because we need to map partitions between topics and brokers
        # So we do shallow copies and populate the partition information ourselves
        for broker in self.brokers:
            newcluster.add_broker(self.brokers[broker].copy())

        for topic_name in self.topics:
            topic = self.topics[topic_name]
            newtopic = topic.copy()

            for i, partition in enumerate(topic.partitions):
                newpartition = partition.copy()
                newtopic.add_partition(newpartition)
                for pos, broker in enumerate(partition.replicas):
                    newpartition.add_replica(newcluster.brokers[broker.id])

            newcluster.add_topic(newtopic)

        return newcluster

    def add_broker(self, broker):
        broker.cluster = self
        self.brokers[broker.id] = broker

    def add_topic(self, topic):
        topic.cluster = self
        topic.retention = self.retention
        self.topics[topic.name] = topic

    def add_group(self, group):
        group.cluster = self
        self.groups[group.name] = group

    # Iterate over all the partitions in this cluster
    # Order is alphabetical by topic, numeric by partition
    def partitions(self, exclude_topics=[]):
        for topic in sorted(self.topics):
            if topic in exclude_topics:
                log.debug("Skipping topic {0} due to exclude-topics".format(topic))
                continue

            for partition in self.topics[topic].partitions:
                yield partition

    def num_brokers(self):
        return len(self.brokers)

    def num_topics(self):
        return len(self.topics)

    def max_replication_factor(self):
        max_pos = 0
        for broker in self.brokers:
            broker_max_pos = 0
            if len(self.brokers[broker].partitions) > 0:
                broker_max_pos = max(self.brokers[broker].partitions.keys())
            if broker_max_pos > max_pos:
                max_pos = broker_max_pos
        return max_pos + 1

    def changed_partitions(self, newcluster):
        moves = []
        try:
            for partition in newcluster.partitions([]):
                if partition.replicas != self.topics[partition.topic.name].partitions[partition.num].replicas:
                    moves.append(partition)
        except KeyError:
            raise ClusterConsistencyException("The new cluster does not match the existing cluster, either due to added topics or partitions")
        return moves

    def log_broker_summary(self):
        for broker_id in sorted(self.brokers.keys()):
            broker = self.brokers[broker_id]
            log.info("Broker {0}: partitions={1}/{2} ({3:.2f}%), size={4}".format(broker_id,
                                                                                  broker.num_leaders(),
                                                                                  broker.num_partitions(),
                                                                                  broker.percent_leaders(),
                                                                                  broker.total_size()))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import os
import sys
import time

import kafka.tools.assigner.actions
import kafka.tools.assigner.plugins
import kafka.tools.assigner.sizers
from kafka.tools import log
from kafka.tools.utilities import get_tools_path, check_java_home
from kafka.tools.models.cluster import Cluster

from kafka.tools.assigner.arguments import set_up_arguments
from kafka.tools.assigner.batcher import split_partitions_into_batches
from kafka.tools.exceptions import ProgrammingException
from kafka.tools.modules import get_modules
from kafka.tools.assigner.models.reassignment import Reassignment
from kafka.tools.assigner.models.replica_election import ReplicaElection


def get_module_map(base_module_name, base_module_class):
    return dict((cls.name, cls) for cls in get_modules(base_module_name, base_module_class))


def get_plugins_list():
    return get_modules(kafka.tools.assigner.plugins, kafka.tools.assigner.plugins.PluginModule)


def check_and_get_sizes(action_cls, args, cluster, sizer_map):
    if action_cls.needs_sizes:
        sizer_to_run = sizer_map[args.sizer](args, cluster)
        sizer_to_run.get_partition_sizes()

        if args.size:
            log.info("Partition Sizes:")
            for topic in cluster.topics:
                for partition in cluster.topics[topic].partitions:
                    log.info("{0} {1}:{2}".format(partition.size, topic, partition.num))


def run_preferred_replica_elections(batches, args, tools_path, plugins, dry_run):
    for i, batch in enumerate(batches):
        # Sleep between PLEs
        if i > 0 and not dry_run:
            log.info("Waiting {0} seconds for replica election to complete".format(args.ple_wait))
            time.sleep(args.ple_wait)

        log.info("Executing preferred replica election {0}/{1}".format(i + 1, len(batches)))
        batch.execute(i + 1, len(batches), args.zookeeper, tools_path, plugins, dry_run)


def get_all_plugins():
    return [plugin() for plugin in get_plugins_list()]


def run_plugins_at_step(plugins, step_name, *args):
    for plugin in plugins:
        try:
            func = getattr(plugin, step_name)
        except AttributeError:
            raise ProgrammingException("Attempt to call plugins with an unknown step")
        if len(args) > 0:
            func(args)
        else:
            func()


def print_leadership(type_str, cluster, dont_skip):
    if dont_skip:
        log.info("Cluster Leadership Balance (before):")
        cluster.log_broker_summary()


def is_dry_run(args):
    if args.generate or not args.execute:
        log.info("--execute flag NOT specified. DRY RUN ONLY")
        return True
    return False


def main():
    # Start by loading all the modules
    action_map = get_module_map(kafka.tools.assigner.actions, kafka.tools.assigner.actions.ActionModule)
    sizer_map = get_module_map(kafka.tools.assigner.sizers, kafka.tools.assigner.sizers.SizerModule)
    plugins = get_all_plugins()

    # Set up and parse all CLI arguments
    args = set_up_arguments(action_map, sizer_map, plugins)
    run_plugins_at_step(plugins, 'set_arguments', args)

    tools_path = get_tools_path(args.tools_path)
    check_java_home()

    cluster = Cluster.create_from_zookeeper(args.zookeeper, getattr(args, 'default_retention', 1))
    run_plugins_at_step(plugins, 'set_cluster', cluster)

    # If the module needs the partition sizes, call a size module to get the information
    check_and_get_sizes(action_map[args.action], args, cluster, sizer_map)
    run_plugins_at_step(plugins, 'after_sizes')
    print_leadership("before", cluster, args.leadership)

    # Clone the cluster, and run the action to generate a new cluster state
    newcluster = cluster.clone()
    action_to_run = action_map[args.action](args, newcluster)
    action_to_run.process_cluster()
    run_plugins_at_step(plugins, 'set_new_cluster', action_to_run.cluster)
    print_leadership("after", newcluster, args.leadership)

    move_partitions = cluster.changed_partitions(action_to_run.cluster)
    batches = split_partitions_into_batches(move_partitions, batch_size=args.moves, use_class=Reassignment)
    run_plugins_at_step(plugins, 'set_batches', batches)

    log.info("Partition moves required: {0}".format(len(move_partitions)))
    log.info("Number of batches: {0}".format(len(batches)))
    dry_run = is_dry_run(args)

    for i, batch in enumerate(batches):
        log.info("Executing partition reassignment {0}/{1}: {2}".format(i + 1, len(batches), repr(batch)))
        batch.execute(i + 1, len(batches), args.zookeeper, tools_path, plugins, dry_run)

    run_plugins_at_step(plugins, 'before_ple')

    if not args.skip_ple:
        all_cluster_partitions = [p for p in action_to_run.cluster.partitions(args.exclude_topics)]
        batches = split_partitions_into_batches(all_cluster_partitions, batch_size=args.ple_size, use_class=ReplicaElection)
        log.info("Number of replica elections: {0}".format(len(batches)))
        run_preferred_replica_elections(batches, args, tools_path, plugins, dry_run)

    run_plugins_at_step(plugins, 'finished')

    return os.EX_OK


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import random
from collections import deque
from operator import attrgetter

from kafka.tools import log
from kafka.tools.exceptions import BalanceException
from kafka.tools.assigner.actions import ActionBalanceModule


class ActionBalanceRackAware(ActionBalanceModule):
    name = "rackaware"
    helpstr = "Reassign partition replicas to assure they are in different racks, attempting to maintain counts and size per-broker"

    def __init__(self, args, cluster):
        super(ActionBalanceRackAware, self).__init__(args, cluster)

        # Set up a random ordered deque of brokers to use when we can't swap
        broker_list = list(self.cluster.brokers.values())
        random.shuffle(broker_list)
        self._random_brokers = deque(broker_list)

    def process_cluster(self):
        log.info("Starting partition balance by rack")

        # Check if rack information is set for the cluster
        broker_racks = [broker.rack for broker in self.cluster.brokers.values()]
        if len(set(broker_racks)) == 1:
            raise BalanceException("Cannot balance cluster by rack as it has no rack information")

        # Figure out the max RF for the cluster
        max_rf = self.cluster.max_replication_factor()

        # Balance partitions at each position separately
        for pos in range(max_rf):
            self._process_partitions_at_pos(pos)

    def _get_sorted_partition_list_at_pos(self, pos):
        # Create a list of partitions to use at this position, sorted by size
        partitions = [p for p in self.cluster.partitions(self.args.exclude_topics) if (len(p.replicas) > (pos + 1))]
        partitions.sort(key=attrgetter('size'))
        return partitions

    def _process_partitions_at_pos(self, pos):
        # Create a sorted list of partitions to use at this position (descending size)
        partitions = self._get_sorted_partition_list_at_pos(pos)

        for i, partition in enumerate(partitions):
            # Check if the partition is already on different racks
            partition_racks = racks_for_replica_list(partition.replicas)
            if len(partition_racks) == len(set(partition_racks)):
                continue

            # Find the partition that is closest in size that has a replica that can be swapped
            # We slice the partitions list here, leaving off the partition we're working on. This way we can just pop off the lists
            large_partitions = partitions[min(i+1, len(partitions)):]
            large_partitions.reverse()
            swap_partition = self._try_pick_swap_partition(partition, pos, partitions[0:i], large_partitions)

            if swap_partition is not None:
                # Swap the replicas at the current position on both partitions
                orig_replica = partition.replicas[pos]
                partition.swap_replicas(orig_replica, swap_partition.replicas[pos])
                swap_partition.swap_replicas(swap_partition.replicas[pos], orig_replica)
            else:
                # If we don't have a swap, pick the next broker in a different rack
                swap_broker = self._try_pick_new_broker(partition, pos)
                partition.swap_replicas(partition.replicas[pos], swap_broker)

    def _try_pick_swap_partition(self, partition, pos, small_partitions, large_partitions):
        """
        Given a partition and a replica position, try and find a partition that has a replica in the same position
        that can be swapped to improve rack separation. The partition selected should be the closest in size possible.

        :params partiton: the Partition with the replica that needs to be changed
        :params pos: the position of the replica to be changed
        :params small_partitions: a list of Partition objects, sorted in order by size, smaller than the Partion to change
        :params large_partitions: a list of Partition objects, sorted in order by size, larger than the Partion to change
        :returns: a Partition object that is OK to swap, or None
        """
        while len(small_partitions) + len(large_partitions) > 0:
            smaller_size_diff = difference_in_size_to_last_partition(partition, small_partitions)
            larger_size_diff = difference_in_size_to_last_partition(partition, large_partitions)

            # Check the partition with the smallest size difference
            target = None
            if larger_size_diff < smaller_size_diff:
                target = large_partitions.pop()
            else:
                target = small_partitions.pop()
            if check_partition_swappable(partition.replicas, target.replicas, pos):
                return target

        return None

    def _try_pick_new_broker(self, partition, pos):
        """
        Given a partition and a replica position, pick a new broker for that position that is in a different rack than other replicas

        :params partition: the Partition to modify
        :params pos: the replica position to change
        :returns: the Broker to swap in at the specified position
        :raises: BalanceException if a broker cannot be found that is in a different rack
        """
        replica_racks = set(racks_for_replica_list(partition.replicas, pos))
        for bidx in range(len(self._random_brokers)):
            broker = self._random_brokers.popleft()
            self._random_brokers.append(broker)
            if broker in partition.replicas:
                continue
            if broker.rack not in replica_racks:
                return broker

        raise BalanceException("Cannot find a swap for broker {0} on partition {1}:{2} at position {3}".format(partition.replicas[pos].id,
                                                                                                               partition.topic.name,
                                                                                                               partition.num,
                                                                                                               pos))


def difference_in_size_to_last_partition(partition, partitions):
    """
    Return the difference in size between the specified Partition and the last Partition in the provided list. If the list is
    empty, return infinity.

    :params partition: a Partition object to use for calculating the difference
    :params partitions: a list of Partition objects
    :returns: The difference in size between partition and the last Partition in the partitions list, or infinity
    """
    if len(partitions) == 0:
        return float("inf")
    return abs(partition.size - partitions[-1].size)


def racks_for_replica_list(replicas, pos=None):
    """
    Returns a set of racks for each of the given replicas in the list
    Skip the replica at position pos, if specified

    :params replicas: a list of Broker objects
    :params pos: a replica position to skip, or None to not skip a replica
    :returns: a list of racks
    """
    return [replica.rack for i, replica in enumerate(replicas) if (pos is None) or (i != pos)]


def check_partition_swappable(replicas_a, replicas_b, pos):
    """
    Check if the broker at position pos in the first replica list can be swapped with the replica at position pos in the second list
    1. replicas_a[pos] must not be in the replicas_b list
    2. replicas_b[pos] must not be in the replicas_a list
    3. replicas_a[pos] must have a different rack than the replicas in replicas_b (except for replicas_b[pos])
    4. replicas_b[pos] must have a different rack than the replicas in replicas_a (except for replicas_a[pos])

    :params replicas_a: the first replica list
    :params replicas_b: the second replica list
    :params pos: the position in the replica list to be replaced
    :returns: True if the broker can be swapped into this replica list, False otherwise
    """
    if replicas_a[pos] in replicas_b or replicas_b[pos] in replicas_a:
        return False
    return replicas_a[pos].rack not in racks_for_replica_list(replicas_b, pos) and replicas_b[pos].rack not in racks_for_replica_list(replicas_a, pos)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.assigner.actions import ActionModule
from kafka.tools.exceptions import NotEnoughReplicasException


class ActionTrim(ActionModule):
    name = "trim"
    helpstr = "Remove partitions from some brokers (reducing RF)"

    def __init__(self, args, cluster):
        super(ActionTrim, self).__init__(args, cluster)

        self.check_brokers(type_str="Brokers to remove")
        self.brokers = args.brokers

    @classmethod
    def _add_args(cls, parser):
        parser.add_argument('-b', '--brokers', help="List of broker IDs to remove", required=True, type=int, nargs='*')

    def process_cluster(self):
        # For each broker specified, remove it from the replica list for all its partitions
        for broker_id in self.brokers:
            broker = self.cluster.brokers[broker_id]
            for position in broker.partitions:
                partition_list = broker.partitions[position][:]
                for partition in partition_list:
                    self._process_partition(broker, partition)

    def _process_partition(self, broker, partition):
        if partition.topic.name in self.args.exclude_topics:
            return

        partition.remove_replica(broker)
        if len(partition.replicas) < 1:
            raise NotEnoughReplicasException("Cannot trim {0}:{1} as it would result in an empty replica list".format(
                partition.topic.name, partition.num))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from operator import attrgetter

from kafka.tools import log
from kafka.tools.assigner.actions import ActionBalanceModule


class ActionBalanceSize(ActionBalanceModule):
    name = "size"
    helpstr = "Move the largest partitions in the cluster to even the total size on disk per-broker for each replica position"

    # We override this so we can select either the size or the scaled_size attributes to sort on
    def __init__(self, args, cluster, size_attr='size'):
        super(ActionBalanceSize, self).__init__(args, cluster)
        self._size_attr = size_attr

    def process_cluster(self):
        log.info("Starting partition balance by {0}".format(self._size_attr))

        # Figure out the max RF for the cluster
        max_rf = self.cluster.max_replication_factor()

        # Calculate cluster information and sorted partition lists first
        partitions = {}
        sizes = {}
        targets = {}
        margins = {}
        for pos in range(max_rf):
            sizes[pos] = {}
            targets[pos] = {}
            margins[pos] = {}

            # Create a sorted list of partitions to use at this position (descending size)
            # Throw out partitions that are 4K or less in size, as they are effectively empty
            partitions[pos] = [p for p in self.cluster.partitions(self.args.exclude_topics) if (len(p.replicas) > pos) and (getattr(p, self._size_attr) > 4)]
            if len(partitions[pos]) == 0:
                continue
            partitions[pos].sort(key=attrgetter(self._size_attr), reverse=True)

            # Calculate broker size at this position
            for broker in self.cluster.brokers:
                if pos in self.cluster.brokers[broker].partitions:
                    sizes[pos][broker] = sum([getattr(p, self._size_attr) for p in self.cluster.brokers[broker].partitions[pos]], 0)
                else:
                    sizes[pos][broker] = 0

            # Calculate the median size of partitions (margin is median/2) and the average size per broker to target
            # Yes, I know the median calculation is slightly broken (it keeps integers). This is OK
            targets[pos] = sum([getattr(p, self._size_attr) for p in partitions[pos]], 0) // len(self.cluster.brokers)
            sizelen = len(partitions[pos])
            if not sizelen % 2:
                margins[pos] = (getattr(partitions[pos][sizelen // 2], self._size_attr) + getattr(partitions[pos][sizelen // 2 - 1], self._size_attr)) // 4
            else:
                margins[pos] = getattr(partitions[pos][sizelen // 2], self._size_attr) // 2

        # Balance partitions for each replica position separately
        for pos in range(max_rf):
            if len(sizes[pos]) == 0:
                continue

            log.info("Calculating ideal state for replica position {0}".format(pos))
            log.debug("Target average size per-broker is {0} kibibytes (+/- {1})".format(targets[pos], margins[pos]))

            for broker_id in self.cluster.brokers:
                broker = self.cluster.brokers[broker_id]

                # Skip brokers that are larger than our minimum target size
                min_move = targets[pos] - margins[pos] - sizes[pos][broker_id]
                max_move = min_move + (margins[pos] * 2)
                if min_move <= 0:
                    continue
                log.debug("Moving between {0} and {1} kibibytes to broker {2}".format(min_move, max_move, broker_id))

                # Find partitions to move to this broker
                for partition in partitions[pos]:
                    partition_size = getattr(partition, self._size_attr)

                    # We can use this partition if all of the following are true: the partition has a replica at this position,
                    # it's size is less than or equal to the max move size, the broker at this replica position would not go out
                    # of range, and it doesn't already exist on this broker at this position
                    if ((len(partition.replicas) <= pos) or (partition_size > max_move) or
                       ((sizes[pos][partition.replicas[pos].id] - partition_size) < (targets[pos] - margins[pos])) or
                       (partition.replicas[pos] == broker)):
                        continue

                    # We can only use a partition that this replica exists on if swapping positions wouldn't hurt balance of the other position or broker
                    source = partition.replicas[pos]
                    if broker in partition.replicas:
                        other_pos = partition.replicas.index(broker)
                        if ((sizes[other_pos][broker_id] - partition_size < targets[other_pos] - margins[other_pos]) or
                           (sizes[other_pos][source.id] + partition_size > targets[pos] + margins[pos]) or
                           (sizes[pos][broker_id] + partition_size > targets[pos] + margins[pos]) or
                           (sizes[pos][source.id] - partition_size < targets[pos] - margins[pos])):
                            continue

                        partition.swap_replica_positions(source, broker)
                        sizes[other_pos][broker_id] -= partition_size
                        sizes[other_pos][source.id] += partition_size
                    else:
                        # Move the partition and adjust sizes
                        partition.swap_replicas(source, broker)
                    sizes[pos][broker_id] += partition_size
                    sizes[pos][source.id] -= partition_size
                    min_move -= partition_size
                    max_move -= partition_size

                    # If we have moved enough partitions, stop for this broker
                    if min_move <= 0:
                        break



import unittest
from mock import patch, MagicMock

from tests.tools.client.fixtures import group_coordinator, group_coordinator_error

from kafka.tools.client import Client
from kafka.tools.exceptions import GroupError, ConnectionError
from kafka.tools.models.broker import Broker


class SendHelperTests(unittest.TestCase):
    def setUp(self):
        # Dummy client for testing - we're not going to connect that bootstrap broker
        self.client = Client()
        self.group_coordinator = group_coordinator()
        self.coordinator_error = group_coordinator_error()

    def test_send_group_aware_request(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker1.rack = 'rack1'
        broker1.send = MagicMock()
        broker1.send.return_value = (1, 'fakeresponse')
        self.client.cluster.add_broker(broker1)

        self.client._send_any_broker = MagicMock()
        self.client._send_any_broker.return_value = self.group_coordinator
        self.client._send_group_aware_request('testgroup', 'fakerequest')

        assert 'testgroup' in self.client.cluster.groups
        assert self.client.cluster.groups['testgroup'].coordinator == broker1
        broker1.send.assert_called_once_with('fakerequest')

    @patch('kafka.tools.client.Broker')
    def test_send_group_aware_request_new_broker(self, mock_broker_class):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker1.rack = 'rack1'
        broker1.send = MagicMock()
        broker1.send.return_value = (1, 'fakeresponse')
        mock_broker_class.return_value = broker1

        self.client._send_any_broker = MagicMock()
        self.client._send_any_broker.return_value = self.group_coordinator
        self.client._send_group_aware_request('testgroup', 'fakerequest')

        mock_broker_class.assert_called_once_with('host1.example.com', id=1, port=8031)
        assert 1 in self.client.cluster.brokers
        assert self.client.cluster.brokers[1] == broker1
        assert self.client.cluster.groups['testgroup'].coordinator == broker1
        broker1.send.assert_called_once_with('fakerequest')

    def test_send_group_aware_request_error(self):
        self.client._send_any_broker = MagicMock()
        self.client._send_any_broker.return_value = self.coordinator_error
        self.assertRaises(GroupError, self.client._send_group_aware_request, 'testgroup', 'fakerequest')

    def test_send_all_brokers(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker1.rack = 'rack1'
        broker1.send = MagicMock()
        broker1.send.return_value = (1, 'fakeresponse')
        broker2 = Broker('host2.example.com', id=101, port=8032)
        broker2.rack = 'rack1'
        broker2.send = MagicMock()
        broker2.send.return_value = (2, 'otherresponse')
        self.client.cluster.add_broker(broker1)
        self.client.cluster.add_broker(broker2)

        val = self.client._send_all_brokers('fakerequest')
        broker1.send.assert_called_once_with('fakerequest')
        broker2.send.assert_called_once_with('fakerequest')
        assert val[1] == 'fakeresponse'
        assert val[101] == 'otherresponse'

    def test_send_all_brokers_error(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker1.rack = 'rack1'
        broker1.send = MagicMock()
        broker1.send.return_value = (1, 'fakeresponse')
        broker2 = Broker('host2.example.com', id=101, port=8032)
        broker2.rack = 'rack1'
        broker2.send = MagicMock()
        broker2.send.side_effect = ConnectionError
        self.client.cluster.add_broker(broker1)
        self.client.cluster.add_broker(broker2)

        val = self.client._send_all_brokers('fakerequest')
        broker1.send.assert_called_once_with('fakerequest')
        broker2.send.assert_called_once_with('fakerequest')
        assert val[1] == 'fakeresponse'
        assert val[101] is None

    @patch('kafka.tools.client.shuffle', lambda x: sorted(x))
    def test_send_any_broker(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker1.rack = 'rack1'
        broker1.send = MagicMock()
        broker1.send.return_value = (1, 'fakeresponse')
        broker2 = Broker('host2.example.com', id=101, port=8032)
        broker2.rack = 'rack1'
        broker2.send = MagicMock()
        broker2.send.return_value = (2, 'otherresponse')
        self.client.cluster.add_broker(broker1)
        self.client.cluster.add_broker(broker2)

        val = self.client._send_any_broker('fakerequest')
        broker2.send.assert_called_once_with('fakerequest')
        assert val == 'otherresponse'

    @patch('kafka.tools.client.shuffle', lambda x: sorted(x))
    def test_send_any_broker_one_error(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker1.rack = 'rack1'
        broker1.send = MagicMock()
        broker1.send.return_value = (1, 'fakeresponse')
        broker2 = Broker('host2.example.com', id=101, port=8032)
        broker2.rack = 'rack1'
        broker2.send = MagicMock()
        broker2.send.side_effect = ConnectionError
        self.client.cluster.add_broker(broker1)
        self.client.cluster.add_broker(broker2)

        val = self.client._send_any_broker('fakerequest')
        broker2.send.assert_called_once_with('fakerequest')
        broker1.send.assert_called_once_with('fakerequest')
        assert val == 'fakeresponse'

    @patch('kafka.tools.client.shuffle', lambda x: sorted(x))
    def test_send_all_error(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker1.rack = 'rack1'
        broker1.send = MagicMock()
        broker1.send.side_effect = ConnectionError
        broker2 = Broker('host2.example.com', id=101, port=8032)
        broker2.rack = 'rack1'
        broker2.send = MagicMock()
        broker2.send.side_effect = ConnectionError
        self.client.cluster.add_broker(broker1)
        self.client.cluster.add_broker(broker2)

        self.assertRaises(ConnectionError, self.client._send_any_broker, 'fakerequest')
        broker1.send.assert_called_once_with('fakerequest')
        broker2.send.assert_called_once_with('fakerequest')



import sys
import unittest

from argparse import Namespace
from .fixtures import set_up_cluster, set_up_subparser

from kafka.tools.exceptions import ConfigurationException, NotEnoughReplicasException
from kafka.tools.assigner.actions.remove import ActionRemove
from kafka.tools.models.broker import Broker


class ActionRemoveTests(unittest.TestCase):
    def setUp(self):
        self.cluster = set_up_cluster()
        (self.parser, self.subparsers) = set_up_subparser()
        self.args = Namespace(exclude_topics=[])

    def test_create_class(self):
        self.args.brokers = [1]
        self.args.to_brokers = [2]
        action = ActionRemove(self.args, self.cluster)
        assert isinstance(action, ActionRemove)

    def test_create_class_bad_target(self):
        self.args.brokers = [1]
        self.args.to_brokers = [3]
        self.assertRaises(ConfigurationException, ActionRemove, self.args, self.cluster)

    def test_create_class_bad_source(self):
        self.args.brokers = [3]
        self.args.to_brokers = [2]
        self.assertRaises(ConfigurationException, ActionRemove, self.args, self.cluster)

    def test_create_class_source_in_target(self):
        self.args.brokers = [1]
        self.args.to_brokers = [1]
        self.assertRaises(ConfigurationException, ActionRemove, self.args, self.cluster)

    def test_create_class_no_target(self):
        self.args.brokers = [1]
        self.args.to_brokers = []
        action = ActionRemove(self.args, self.cluster)
        assert action.to_brokers == [2]

    def test_configure_args(self):
        ActionRemove.configure_args(self.subparsers)
        sys.argv = ['kafka-assigner', 'remove', '-b', '1', '-t', '2']
        parsed_args = self.parser.parse_args()
        assert parsed_args.action == 'remove'

    def test_process_cluster(self):
        self.cluster.add_broker(Broker("brokerhost3.example.com", id=3))
        self.args.brokers = [1]
        self.args.to_brokers = []
        action = ActionRemove(self.args, self.cluster)
        action.process_cluster()

        b2 = self.cluster.brokers[2]
        b3 = self.cluster.brokers[3]
        assert self.cluster.topics['testTopic1'].partitions[0].replicas == [b3, b2]
        assert self.cluster.topics['testTopic1'].partitions[1].replicas == [b2, b3]
        assert self.cluster.topics['testTopic2'].partitions[0].replicas == [b2, b3]
        assert self.cluster.topics['testTopic2'].partitions[1].replicas == [b3, b2]

    def test_process_cluster_exclude(self):
        self.cluster.add_broker(Broker("brokerhost3.example.com", id=3))
        self.args.brokers = [1]
        self.args.to_brokers = []
        self.args.exclude_topics = ['testTopic1']
        action = ActionRemove(self.args, self.cluster)
        action.process_cluster()

        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        b3 = self.cluster.brokers[3]
        assert self.cluster.topics['testTopic1'].partitions[0].replicas == [b1, b2]
        assert self.cluster.topics['testTopic1'].partitions[1].replicas == [b2, b1]
        assert self.cluster.topics['testTopic2'].partitions[0].replicas == [b2, b3]
        assert self.cluster.topics['testTopic2'].partitions[1].replicas == [b3, b2]

    def test_process_cluster_error(self):
        self.args.brokers = [1]
        self.args.to_brokers = []
        action = ActionRemove(self.args, self.cluster)
        self.assertRaises(NotEnoughReplicasException, action.process_cluster)



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.offset_commit_v0 import OffsetCommitV0Request


class OffsetCommitV0RequestTests(unittest.TestCase):
    def test_process_arguments(self):
        val = OffsetCommitV0Request.process_arguments(['groupname', 'topicname', '4,2,somemetadata', 'nexttopic', '9,3,moremetadata'])
        assert val == {'group_id': 'groupname',
                       'topics': [{'topic': 'topicname', 'partitions': [{'partition': 4, 'offset': 2, 'metadata': 'somemetadata'}]},
                                  {'topic': 'nexttopic', 'partitions': [{'partition': 9, 'offset': 3, 'metadata': 'moremetadata'}]}]}

    def test_process_arguments_notenough(self):
        self.assertRaises(ArgumentError, OffsetCommitV0Request.process_arguments, ['groupname', 'topicname'])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import collections
import six

from kafka.tools.protocol.types import BaseType
from kafka.tools.protocol.types.boolean import Boolean
from kafka.tools.protocol.types.bytes import Bytes
from kafka.tools.protocol.types.integers import Int8, Int16, Int32, Int64
from kafka.tools.protocol.types.string import String


# We need to define this early, but Array won't be in scope until later
schema_type_map = {}


def indent(text, amount, ch=' '):
    padding = amount * ch
    return ''.join(padding+line for line in text.splitlines(True))


def _validate_schema_entry_type(entry_type):
    if isinstance(entry_type, six.string_types):
        if entry_type.lower() not in schema_type_map:
            raise TypeError('Unknown schema data type: {0}'.format(entry_type))
    elif isinstance(entry_type, collections.Sequence):
        _validate_schema(entry_type)
    else:
        raise TypeError('Schema type must be a string type or a list, not {0}'.format(type(entry_type)))


def _validate_schema_entry(entry):
    if not isinstance(entry, dict):
        raise TypeError('Expected schema entries to be dicts, not {0}'.format(type(entry)))
    if not all(k in entry for k in ('name', 'type')):
        raise KeyError('Schema entries must have both name and type keys')
    if not isinstance(entry['name'], six.string_types):
        raise TypeError('Schema entry name must be a string, not {0}'.format(type(entry['name'])))


def _validate_schema(schema):
    for entry in schema:
        _validate_schema_entry(entry)
        if isinstance(entry['type'], six.string_types) and (entry['type'].lower() == 'array'):
            if 'item_type' not in entry:
                raise KeyError('Array schema entries must have an item_type key')
            _validate_schema_entry_type(entry['item_type'])
        else:
            _validate_schema_entry_type(entry['type'])


def _evaluate_value(value, schema):
    rv = {}
    for entry in schema:
        if entry['name'] not in value:
            raise KeyError('Value is missing an entry with key "{0}"'.format(entry['name']))
        entry_schema = entry['type']
        if isinstance(entry_schema, six.string_types):
            if entry_schema.lower() == 'array':
                entry_schema = entry['item_type']
            rv[entry['name']] = schema_type_map[entry['type'].lower()](value[entry['name']], entry_schema)
        else:
            rv[entry['name']] = Sequence(value[entry['name']], entry_schema)
    return rv


class SequenceIterator():
    def __init__(self, iter_obj, reversed=False):
        self._iter_obj = iter_obj
        self._reversed = reversed

        if self._reversed:
            # We're going to use negative indexing for reverse
            self._next_idx = 1
            self._last_idx = len(self._iter_obj._schema)
        else:
            self._next_idx = 0
            self._last_idx = len(self._iter_obj._schema) - 1

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        if self._next_idx > self._last_idx:
            raise StopIteration
        rv_idx = self._next_idx
        self._next_idx += 1
        if self._reversed:
            return self._iter_obj[self._iter_obj._schema[-rv_idx]['name']]
        else:
            return self._iter_obj[self._iter_obj._schema[rv_idx]['name']]


class Sequence(BaseType):
    _type = 'sequence'

    def _validate(self):
        if (not isinstance(self._schema, collections.Sequence)) or isinstance(self._schema, six.string_types):
            raise TypeError('Expected schema to be a list, not {0}'.format(type(self._schema)))
        _validate_schema(self._schema)

        if not isinstance(self._value, dict):
            raise TypeError('Expected value to be a dict, not {0}'.format(type(self._value)))
        self._value = _evaluate_value(self._value, self._schema)

    @classmethod
    def decode(cls, byte_array, schema=None):
        _validate_schema(schema)

        val = {}
        remaining_bytes = byte_array
        for entry in schema:
            if entry['type'].lower() == 'array':
                val[entry['name']], remaining_bytes = schema_type_map[entry['type']].decode(remaining_bytes, schema=entry['item_type'])
            else:
                val[entry['name']], remaining_bytes = schema_type_map[entry['type']].decode(remaining_bytes)

        return cls(val, schema=schema, skip_validation=True), remaining_bytes

    def encode(self):
        byte_array = b''
        for item in self._schema:
            byte_array += self._value[item['name']].encode()
        return byte_array

    def __len__(self):
        return len(self._value)

    def __contains__(self, k):
        return k in self._value

    def __getitem__(self, k):
        return self._value[k]

    def __setitem__(self, k, value):
        raise NotImplementedError

    def __delitem__(self, k):
        raise NotImplementedError

    def __iter__(self):
        return SequenceIterator(self)

    def __reversed__(self):
        return SequenceIterator(self, reversed=True)

    def __str__(self):
        lines = ["{0}: {1}".format(item['name'], self._value[item['name']]) for item in self._schema]
        return "\n".join(lines)

    def __repr__(self):
        return '<Sequence length={0}>'.format(len(self))


class ArrayIterator():
    def __init__(self, iter_obj, reversed=False):
        self._iter_obj = iter_obj
        self._reversed = reversed

        if self._reversed:
            # We're going to use negative indexing for reverse
            self._next_idx = 1
            self._last_idx = len(self._iter_obj)
        else:
            self._next_idx = 0
            self._last_idx = len(self._iter_obj) - 1

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        if self._next_idx > self._last_idx:
            raise StopIteration
        rv_idx = self._next_idx
        self._next_idx += 1
        if self._reversed:
            return self._iter_obj[-rv_idx]
        else:
            return self._iter_obj[rv_idx]


class Array(BaseType):
    _type = "array"

    def _evaluate_value(self):
        try:
            klass = schema_type_map[self._schema.lower()] if isinstance(self._schema, six.string_types) else Sequence
        except KeyError:
            raise TypeError('Unknown schema type {0}'.format(self._schema.lower()))
        self._value = [klass(item, schema=self._schema) for item in self._value]

    def _validate(self):
        if self._value is None:
            return
        if (not isinstance(self._value, collections.Sequence)) or isinstance(self._value, six.string_types):
            raise ValueError("The value must be None or a list, not {0}".format(type(self._value)))
        self._evaluate_value()

    @classmethod
    def decode(cls, byte_array, schema=None):
        array_len, remaining_bytes = Int32.decode(byte_array)
        if array_len.value() == -1:
            return cls(None, schema=schema), remaining_bytes

        try:
            klass = schema_type_map[schema.lower()] if isinstance(schema, six.string_types) else Sequence
        except KeyError:
            raise TypeError('Unknown schema type {0}'.format(schema.lower()))
        val = []
        for i in range(array_len.value()):
            item, remaining_bytes = klass.decode(remaining_bytes, schema)
            val.append(item)
        return cls(val, schema=schema, skip_validation=True), remaining_bytes

    def encode(self):
        if self._value is None:
            return Int32(-1).encode()

        byte_array = Int32(len(self._value)).encode()
        for item in self._value:
            byte_array += item.encode()
        return byte_array

    def __len__(self):
        # For the sake of interactions with code that treats this like an iterator, a null array is of zero length
        if self._value is None:
            return 0
        return len(self._value)

    def __contains__(self, item):
        if self._value is None:
            return False
        return any(item == x for x in self._value)

    def __getitem__(self, k):
        return self._value[k]

    def __setitem__(self, k, value):
        raise NotImplementedError

    def __delitem__(self, k):
        raise NotImplementedError

    def __iter__(self):
        return ArrayIterator(self)

    def __reversed__(self):
        return ArrayIterator(self, reversed=True)

    def __str__(self):
        lines = ["{0} items (array)".format(len(self))]
        lines.append("\n    ---\n".join([indent(str(item), 4) for item in self]))
        return "\n".join(lines)

    def __repr__(self):
        return "<Array length={0}>".format(len(self))


# Now we can fully declare this map
schema_type_map = {
    'array': Array,
    'boolean': Boolean,
    'bytes': Bytes,
    'int8': Int8,
    'int16': Int16,
    'int32': Int32,
    'int64': Int64,
    'string': String,
}



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import glob
import importlib
import inspect
import os


def is_class(klass):
    return inspect.isclass(klass) and (not inspect.isabstract(klass))


def check_class(klass, base_class):
    if not is_class(klass):
        return None
    base_classes = inspect.getmro(klass)
    if len(base_classes) == 1:
        return None
    if base_classes[1] == base_class:
        return klass


def check_classes_in_module(module, base_class):
    module_list = []
    for item in dir(module):
        value = getattr(module, item)
        if check_class(value, base_class) is not None:
            module_list.append(value)

    return module_list


def get_modules(base_package, base_class):
    module_list = []
    module_file_paths = glob.glob(os.path.join(base_package.__path__[0], "*.py"))
    for module_file_path in module_file_paths:
        module_filename = os.path.basename(module_file_path)
        module_name = os.path.splitext(module_filename)[0]
        if module_name.startswith("__"):
            continue

        # Import the module
        module = importlib.import_module("." + module_name, package=base_package.__name__)

        # Add all children of the base_class in the module
        module_list.extend(check_classes_in_module(module, base_class))

    return list(set(module_list))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.protocol.requests import BaseRequest
from kafka.tools.protocol.responses.update_metadata_v0 import UpdateMetadataV0Response
from kafka.tools.protocol.requests.leader_and_isr_v0 import _process_arguments


class UpdateMetadataV0Request(BaseRequest):
    api_key = 6
    api_version = 0
    cmd = "UpdateMetadata"
    response = UpdateMetadataV0Response

    help_string = ("Request:     {0}V{1}\n".format(cmd, api_version) +
                   "Format:      {0}V{1} controller_id controller_epoch ".format(cmd, api_version) +
                   "(topic,partition,controller_epoch,leader,leader_epoch,isr,zk_version,replicas ...) (broker_id,host,port ...)\n" +
                   "             isr and replicas are a '|' separated list of broker IDs (e.g. '2|3')\n" +
                   "Description: Send updated metadata information to broker\n")

    schema = [
        {'name': 'controller_id', 'type': 'int32'},
        {'name': 'controller_epoch', 'type': 'int32'},
        {'name': 'partition_states',
         'type': 'array',
         'item_type': [
             {'name': 'topic', 'type': 'string'},
             {'name': 'partition', 'type': 'int32'},
             {'name': 'controller_epoch', 'type': 'int32'},
             {'name': 'leader', 'type': 'int32'},
             {'name': 'leader_epoch', 'type': 'int32'},
             {'name': 'isr', 'type': 'array', 'item_type': 'int32'},
             {'name': 'zk_version', 'type': 'int32'},
             {'name': 'replicas', 'type': 'array', 'item_type': 'int32'},
         ]},
        {'name': 'live_leaders',
         'type': 'array',
         'item_type': [
             {'name': 'id', 'type': 'int32'},
             {'name': 'host', 'type': 'string'},
             {'name': 'port', 'type': 'int32'},
         ]},
    ]

    @classmethod
    def process_arguments(cls, cmd_args):
        return _process_arguments("UpdateMetadataV0", cmd_args)



import os
import threading
import time
import unittest

from mock import patch

from kafka.tools.utilities import is_exec_file, get_tools_path, check_java_home, find_path_containing, json_loads, synchronized
from kafka.tools.exceptions import ConfigurationException


class ToolsTests(unittest.TestCase):
    @patch('kafka.tools.utilities.os.path.isfile')
    @patch('kafka.tools.utilities.os.access')
    def test_is_exec_file(self, mock_access, mock_isfile):
        mock_access.return_value = True
        mock_isfile.return_value = True
        fname = 'somefile'
        res = is_exec_file(fname)
        assert res is True

    @patch('kafka.tools.utilities.os.path.isfile')
    @patch('kafka.tools.utilities.os.access')
    def test_is_exec_file_nonexistent(self, mock_access, mock_isfile):
        mock_access.return_value = True
        mock_isfile.return_value = False
        fname = 'this_file_does_not_exist_ksdjkfsj'
        res = is_exec_file(fname)
        assert res is False

    @patch('kafka.tools.utilities.os.path.isfile')
    @patch('kafka.tools.utilities.os.access')
    def test_is_exec_file_noexec(self, mock_access, mock_isfile):
        mock_access.return_value = False
        mock_isfile.return_value = True
        fname = 'this_file_does_not_exist_ksdjkfsj'
        res = is_exec_file(fname)
        assert res is False

    @patch('kafka.tools.utilities.find_path_containing')
    def test_get_tools_path_default(self, mock_find):
        mock_find.return_value = '/path/to/tools'
        tools_path = get_tools_path()
        assert tools_path == '/path/to/tools'

    @patch('kafka.tools.utilities.is_exec_file')
    def test_find_path_containing_found(self, mock_is_exec_file):
        mock_is_exec_file.return_value = True
        tools_path = find_path_containing('some_filename')
        assert tools_path is not None

    @patch('kafka.tools.utilities.is_exec_file')
    def test_find_path_containing_notfound(self, mock_is_exec_file):
        mock_is_exec_file.return_value = False
        self.assertRaises(ConfigurationException, find_path_containing, 'some_filename')

    @patch('kafka.tools.utilities.is_exec_file')
    def test_get_tools_path_explicit_found(self, mock_is_exec_file):
        mock_is_exec_file.return_value = True
        tools_path = get_tools_path('/path/to/file')
        assert tools_path is '/path/to/file'

    @patch('kafka.tools.utilities.is_exec_file')
    def test_get_tools_path_explicit_notfound(self, mock_is_exec_file):
        mock_is_exec_file.return_value = False
        self.assertRaises(ConfigurationException, get_tools_path, '/path/to/file')

    @patch('kafka.tools.utilities.is_exec_file')
    def test_check_java_home_found(self, mock_is_exec_file):
        os.environ['JAVA_HOME'] = '/path/to/java'
        mock_is_exec_file.return_value = True
        check_java_home()
        mock_is_exec_file.assert_called_with('/path/to/java/bin/java')

    @patch('kafka.tools.utilities.is_exec_file')
    def test_check_java_home_notfound(self, mock_is_exec_file):
        os.environ['JAVA_HOME'] = '/path/to/java'
        mock_is_exec_file.return_value = False
        self.assertRaises(ConfigurationException, check_java_home)

    def test_check_java_home_notset(self):
        del os.environ['JAVA_HOME']
        self.assertRaises(ConfigurationException, check_java_home)

    def test_json_loads(self):
        assert json_loads('1') == 1

    def test_json_loads_bytes(self):
        assert json_loads(b'1') == 1

    def test_synchronized(self):
        class TestSyncClass:
            def __init__(self):
                self.calls = []
                self._lock = threading.RLock()

            @synchronized
            def m1(self):
                self.calls.append('m1')
                time.sleep(1)
                self.m2()
                self.calls.append('m1-finish')

            @synchronized
            def m2(self):
                self.calls.append('m2')
                time.sleep(1)
                self.calls.append('m2-finish')

            @synchronized
            def m3(self):
                self.calls.append('m3')
                time.sleep(1)
                self.calls.append('m3-finish')

        def t1(obj):
            obj.m1()

        def t2(obj):
            obj.m3()

        obj = TestSyncClass()
        thread1 = threading.Thread(target=t1, args=(obj,))
        thread2 = threading.Thread(target=t2, args=(obj,))
        thread1.start()
        thread2.start()
        thread1.join(timeout=10.0)
        thread2.join(timeout=10.0)

        assert obj.calls == ['m1', 'm2', 'm2-finish', 'm1-finish', 'm3', 'm3-finish']



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.exceptions import ProgrammingException


def split_partitions_into_batches(partitions, batch_size=10, use_class=None):
    # Currently, this is a very simplistic implementation that just breaks the list of partitions down
    # into even sized chunks. While it could be implemented as a generator, it's not so that it can
    # split the list into more efficient batches.
    if use_class is None:
        raise ProgrammingException("split_partitions_into_batches called with no use_class")

    batches = [use_class(partitions[i:i + batch_size]) for i in range(0, len(partitions), batch_size)]
    return batches



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.controlled_shutdown_v1 import ControlledShutdownV1Request


class ControlledShutdownV1RequestTests(unittest.TestCase):
    def test_process_arguments(self):
        assert ControlledShutdownV1Request.process_arguments([3]) == {'broker_id': 3}

    def test_process_arguments_missing(self):
        self.assertRaises(ArgumentError, ControlledShutdownV1Request.process_arguments, [])

    def test_process_arguments_extra(self):
        self.assertRaises(ArgumentError, ControlledShutdownV1Request.process_arguments, [3, 4])

    def test_process_arguments_nonnumeric(self):
        self.assertRaises(ArgumentError, ControlledShutdownV1Request.process_arguments, ['foo'])



import unittest

from kafka.tools.protocol.requests.topic_metadata_v0 import TopicMetadataV0Request


class TopicMetadataV0RequestTest(unittest.TestCase):
    def test_process_arguments(self):
        val = TopicMetadataV0Request.process_arguments(['topicname', 'anothertopic'])
        assert val == {'topics': ['topicname', 'anothertopic']}

    def test_process_arguments_all(self):
        val = TopicMetadataV0Request.process_arguments([])
        assert val == {'topics': None}



import unittest

from tests.tools.client.fixtures import topic_metadata_v0


class MetadataV0ResponseTest(unittest.TestCase):
    def setUp(self):
        self.metadata = topic_metadata_v0()

    def test_topic_names(self):
        val = self.metadata.topic_names()
        assert val == ['topic1']

    def test_broker_ids(self):
        val = self.metadata.broker_ids()
        assert val == [1, 101]



import unittest

from kafka.tools.protocol.types.bytes import Bytes


class BytesTests(unittest.TestCase):
    def test_from_string(self):
        val = Bytes.from_string("87349DBC")
        assert val.value() == b'\x87\x34\x9d\xbc'

    def test_from_string_lower(self):
        val = Bytes.from_string("87349dbc")
        assert val.value() == b'\x87\x34\x9d\xbc'

    def test_from_string_null(self):
        val = Bytes.from_string(None)
        assert val.value() is None

    def test_from_string_bad_type(self):
        self.assertRaises(TypeError, Bytes.from_string, 123)

    def test_from_string_bad_length(self):
        self.assertRaises(ValueError, Bytes.from_string, "87349DB")

    def test_create_bad_type(self):
        self.assertRaises(TypeError, Bytes, 123)

    def test_decode(self):
        (val, rest) = Bytes.decode(b'\x00\x00\x00\x04\x87\x34\x9d\xbc')
        assert isinstance(val, Bytes)
        assert val.value() == b'\x87\x34\x9d\xbc'
        assert rest == b''

    def test_decode_null(self):
        (val, rest) = Bytes.decode(b'\xff\xff\xff\xff')
        assert isinstance(val, Bytes)
        assert val.value() is None
        assert rest == b''

    def test_decode_remainder(self):
        (val, rest) = Bytes.decode(b'\x00\x00\x00\x04\x87\x34\x9d\xbcabc')
        assert rest == b'abc'

    def test_decode_nosize(self):
        self.assertRaises(ValueError, Bytes.decode, b'')

    def test_decode_underflow(self):
        self.assertRaises(ValueError, Bytes.decode, b'\x00\x00\x00\x04\x874\x9d')

    def test_encode(self):
        val = Bytes(b'\x87\x34\x9d\xbc')
        print(repr(val.encode()))
        assert val.encode() == b'\x00\x00\x00\x04\x87\x34\x9d\xbc'

    def test_encode_null(self):
        val = Bytes(None)
        assert val.encode() == b'\xff\xff\xff\xff'

    def test_str(self):
        val = Bytes(b'\x87\x34\x9d\xbc')
        print(str(val))
        assert str(val) == "87349dbc (bytes)"
        assert repr(val) == "<Bytes length=4>"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import abc
import six

from kafka.tools.protocol.types.sequences import Sequence


class ArgumentError(Exception):
    pass


@six.add_metaclass(abc.ABCMeta)
class BaseRequest():  # pragma: no cover
    @abc.abstractproperty
    def schema(self):
        raise NotImplementedError

    @abc.abstractproperty
    def api_key(self):
        raise NotImplementedError

    @abc.abstractproperty
    def api_version(self):
        pass

    @abc.abstractproperty
    def cmd(self):
        raise NotImplementedError

    @abc.abstractproperty
    def response(self):
        raise NotImplementedError

    @abc.abstractproperty
    def help_string(self):
        raise NotImplementedError

    def __init__(self, value):
        self._request = Sequence(value, schema=self.schema)

    def encode(self):
        return self._request.encode()

    @abc.abstractmethod
    def process_arguments(cls, cmd_args):
        """IMPORTANT: this is class method, override it with @classmethod!"""
        raise NotImplementedError

    def __hash__(self):
        return id(self)

    def __len__(self):
        return len(self._request)

    def __contains__(self, k):
        return k in self._request

    def __getitem__(self, k):
        return self._request[k]

    def __setitem__(self, k):
        raise NotImplementedError

    def __delitem__(self, k):
        raise NotImplementedError



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.exceptions import ConfigurationException


# Base class that all kafka-assigner actions inherit from
class ActionModule(object):
    ##########################################################################################################
    # OVERRIDE ME - These are all the things you can/must override in your child class
    name = ""
    helpstr = ""
    needs_sizes = False

    def __init__(self, args, cluster):
        self.args = args
        self.cluster = cluster

    @classmethod
    def _add_args(cls, parser):
        pass

    def process_cluster(self):
        pass

    ##########################################################################################################
    # MODULE BASICS - You probably shouldn't modify past here unless you're changing the basic way all modules work

    @classmethod
    def configure_args(cls, subparser):
        if cls.name == "" or cls.helpstr == "":
            raise Exception("Cannot instantiate ActionModule directly")

        parser = subparser.add_parser(cls.name, help=cls.helpstr)
        cls._add_args(parser)
        parser.set_defaults(action=cls.name)

    def check_brokers(self, type_str="Source brokers"):
        if len(set(self.args.brokers) & set(self.cluster.brokers)) != len(self.args.brokers):
            raise ConfigurationException("{0} are not in the brokers list for this cluster".format(type_str))


# Special class for balance modules to inherit from that skips configs
class ActionBalanceModule(ActionModule):
    @classmethod
    def configure_args(cls, subparser):
        pass



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.heartbeat_v0 import HeartbeatV0Request


class HeartbeatV0RequestTests(unittest.TestCase):
    def test_process_arguments(self):
        val = HeartbeatV0Request.process_arguments(['groupname', '4', 'membername'])
        assert val == {'group_id': 'groupname', 'group_generation_id': 4, 'member_id': 'membername'}

    def test_process_arguments_missing(self):
        self.assertRaises(ArgumentError, HeartbeatV0Request.process_arguments, [])

    def test_process_arguments_two(self):
        self.assertRaises(ArgumentError, HeartbeatV0Request.process_arguments, ['groupname', 'anothergroup'])

    def test_process_arguments_four(self):
        self.assertRaises(ArgumentError, HeartbeatV0Request.process_arguments, ['groupname', 'foo', 'membername', 'extraarg'])

    def test_process_arguments_nonnumeric(self):
        self.assertRaises(ArgumentError, HeartbeatV0Request.process_arguments, ['groupname', 'foo', 'membername'])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.assigner.actions.balancemodules.size import ActionBalanceSize
from kafka.tools.assigner.actions import ActionBalanceModule


class ActionBalanceRate(ActionBalanceModule):
    name = "rate"
    helpstr = "Move the busiest partitions in the cluster to even the total bytes rate per-broker for each replica position"

    def __init__(self, args, cluster):
        super(ActionBalanceRate, self).__init__(args, cluster)

        # This module merely calls the balance "size" module with the scaled_size attr, so we're going to instantiate a copy of that
        self._size = ActionBalanceSize(args, cluster, size_attr='scaled_size')

    def process_cluster(self):
        # Call the size module
        self._size.process_cluster()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import argparse


def set_up_arguments():
    # Configure basic arguments
    aparser = argparse.ArgumentParser(prog='kafka-protocol', description='CLI tool for sending Kafka protocol requests to a broker')
    aparser.add_argument('-b', '--broker', help='Kafka broker hostname to connect to', default='localhost')
    aparser.add_argument('-p', '--port', help="Port number on the Kafka broker to connect to", type=int, default=9092)
    aparser.add_argument('-t', '--tls', help="Enable TLS for the broker connection", action='store_true')

    # Parse the args that were passed
    args = aparser.parse_args()
    return args



import json
import unittest

from mock import patch, ANY

from kafka.tools.models.broker import Broker
from kafka.tools.models.topic import Topic
from kafka.tools.assigner.models.replica_election import ReplicaElection


class ReplicaElectionTests(unittest.TestCase):
    def setUp(self):
        self.topic = Topic('testTopic', 10)
        self.broker = Broker('brokerhost1.example.com', id=1)
        for i in range(10):
            self.topic.partitions[i].replicas = [self.broker]
        self.replica_election = ReplicaElection(self.topic.partitions, pause_time=0)

    def test_replica_election_create(self):
        assert self.replica_election is not None

    def test_replica_election_repr(self):
        t_repr = json.loads(repr(self.replica_election))
        expect_repr = {'partitions': []}
        for i in range(10):
            expect_repr['partitions'].append({'topic': 'testTopic', 'partition': i})
        assert t_repr == expect_repr

    def test_replica_election_dict(self):
        t_repr = self.replica_election.dict_for_replica_election()
        expect_repr = {'partitions': []}
        for i in range(10):
            expect_repr['partitions'].append({'topic': 'testTopic', 'partition': i})
        assert t_repr == expect_repr

    @patch('kafka.tools.assigner.models.replica_election.subprocess.call')
    def test_replica_election_execute(self, mock_call):
        self.replica_election.execute(1, 1, 'zk_connect_string', '/path/to/tools', plugins=[], dry_run=False)
        mock_call.assert_called_once_with(['/path/to/tools/kafka-preferred-replica-election.sh',
                                           '--zookeeper', 'zk_connect_string',
                                           '--path-to-json-file', ANY],
                                          stderr=ANY, stdout=ANY)



import unittest
from mock import MagicMock, patch, call

from tests.tools.client.fixtures import topic_metadata

from kafka.tools.client import Client
from kafka.tools.configuration import ClientConfiguration
from kafka.tools.exceptions import ConfigurationError, ConnectionError
from kafka.tools.models.broker import Broker
from kafka.tools.models.cluster import Cluster
from kafka.tools.protocol.requests.topic_metadata_v1 import TopicMetadataV1Request


class GenericInterfaceTests(unittest.TestCase):
    def setUp(self):
        # Dummy client for testing - we're not going to connect that bootstrap broker
        self.client = Client(broker_list='broker1.example.com:9091,broker2.example.com:9092')
        self.client._connected = True
        self.metadata_response = topic_metadata()

    def test_close(self):
        # Two brokers for the client
        broker = Broker('host1.example.com', id=1, port=8031)
        broker.rack = 'rack1'
        broker.close = MagicMock()
        self.client.cluster.add_broker(broker)

        broker = Broker('host2.example.com', id=101, port=8032)
        broker.rack = 'rack1'
        broker.close = MagicMock()
        self.client.cluster.add_broker(broker)

        self.client.close()
        for broker_id in self.client.cluster.brokers:
            self.client.cluster.brokers[broker_id].close.assert_called_once()

    def test_create(self):
        assert self.client.configuration.broker_list == [('broker1.example.com', 9091), ('broker2.example.com', 9092)]

    def test_create_with_configuration(self):
        config = ClientConfiguration(zkconnect='zk.example.com:2181/kafka-cluster')
        client = Client(configuration=config)
        assert client.configuration == config

    def test_create_config_zkconnect(self):
        test_client = Client(zkconnect='zk.example.com:2181/kafka-cluster')
        assert test_client.configuration.zkconnect == 'zk.example.com:2181/kafka-cluster'

    def test_create_config_bad_arg(self):
        self.assertRaises(ConfigurationError, Client, invalidconfig='foo')

    def test_create_config_bad_config_object(self):
        self.assertRaises(ConfigurationError, Client, configuration='foo')

    @patch.object(Cluster, 'create_from_zookeeper')
    def test_connect_zookeeper(self, mock_create):
        mock_create.return_value = Cluster()
        test_client = Client(zkconnect='zk.example.com:2181/kafka-cluster')
        test_client.connect()
        mock_create.assert_called_once_with(zkconnect='zk.example.com:2181/kafka-cluster', fetch_topics=False)

    @patch.object(Broker, 'connect')
    @patch.object(Broker, 'send')
    @patch.object(Broker, 'close')
    def test_connect_broker_list(self, mock_close, mock_send, mock_connect):
        def add_brokers_with_mocks(metadata):
            self.client._update_brokers_from_metadata(metadata)
            for broker_id in self.client.cluster.brokers:
                self.client.cluster.brokers[broker_id].connect = MagicMock()

        mock_connect.side_effect = [ConnectionError, None]
        mock_send.return_value = (1, self.metadata_response)
        self.client._update_from_metadata = MagicMock()
        self.client._update_from_metadata.side_effect = add_brokers_with_mocks

        self.client.connect()
        mock_connect.assert_has_calls([call(sslcontext=None), call(sslcontext=None)])
        mock_send.assert_called_once()
        assert isinstance(mock_send.call_args[0][0], TopicMetadataV1Request)
        mock_close.assert_called_once()

        # Metadata is updated and brokers are connected
        self.client._update_from_metadata.assert_called_once_with(self.metadata_response)
        for broker_id in self.client.cluster.brokers:
            self.client.cluster.brokers[broker_id].connect.assert_called_once()

    def test_connect_broker_list_exhausted(self):
        self.client._maybe_bootstrap_cluster = MagicMock()
        self.client._maybe_bootstrap_cluster.return_value = False
        self.assertRaises(ConnectionError, self.client.connect)



import unittest

from argparse import Namespace
from mock import patch, Mock, call

from kafka.tools.exceptions import UnknownBrokerException, ConfigurationException
from kafka.tools.models.broker import Broker
from kafka.tools.models.cluster import Cluster
from kafka.tools.models.topic import Topic
from kafka.tools.assigner.sizers.jmx import SizerJMX, _validate_broker


class MockLong:
    def __init__(self, value):
        self.value = value


class MockMBeanServerConnection:
    def getAttribute(self, bean, attribute):
        return MockLong(1234)

    def queryNames(self, name, query):
        return [MockObjectName("testTopic1", "0")]


class MockJMXConnector:
    def getMBeanServerConnection(self):
        return MockMBeanServerConnection()

    def close(self):
        return


class MockObjectName:
    def __init__(self, topic, partition):
        self.topic = topic
        self.partition = partition

    def getKeyProperty(self, key):
        if key == "topic":
            return self.topic
        if key == "partition":
            return self.partition
        return None


class MockHashMap:
    def __init__(self):
        self._hashmap = {}

    def put(self, key, value):
        self._hashmap[key] = value

    def __getitem__(self, key):
        if key in self._hashmap:
            return self._hashmap[key]
        return None


class SizerJMXTests(unittest.TestCase):
    def setUp(self):
        self.args = Namespace()
        self.args.property = ['libjvm=/path/to/libjvm']
        self.cluster = self.create_cluster_onehost()

        self.mock_jpype = Mock()
        self.mock_hashmap = MockHashMap()
        self.mock_jpype.javax.management.remote.JMXConnectorFactory.connect.return_value = MockJMXConnector()
        self.mock_jpype.javax.management.remote.JMXServiceURL.return_value = "testURL"
        self.mock_jpype.java.util.HashMap.return_value = self.mock_hashmap
        self.mock_jpype.JArray.return_value = list
        self.mock_jpype.javax.management.remote.JMXConnector.CREDENTIALS = "CREDENTIALS"

    def create_cluster_onehost(self):
        cluster = Cluster()
        cluster.add_broker(Broker("brokerhost1.example.com", id=1))
        cluster.brokers[1].jmx_port = 1099
        cluster.add_topic(Topic("testTopic1", 2))
        cluster.add_topic(Topic("testTopic2", 2))
        partition = cluster.topics['testTopic1'].partitions[0]
        partition.add_replica(cluster.brokers[1], 0)
        partition = cluster.topics['testTopic1'].partitions[1]
        partition.add_replica(cluster.brokers[1], 0)
        partition = cluster.topics['testTopic2'].partitions[0]
        partition.add_replica(cluster.brokers[1], 0)
        partition = cluster.topics['testTopic2'].partitions[1]
        partition.add_replica(cluster.brokers[1], 0)
        return cluster

    @patch('kafka.tools.assigner.sizers.jmx.jpype')
    def test_sizer_create(self, mock_jpype_module):
        sizer = SizerJMX(self.args, self.cluster)
        assert isinstance(sizer, SizerJMX)
        mock_jpype_module.startJVM.assert_called_with("/path/to/libjvm")
        mock_jpype_module.java.util.HashMap.assert_called_once()

    @patch('kafka.tools.assigner.sizers.jmx.jpype')
    def test_sizer_create_default(self, mock_jpype_module):
        self.args.property = []
        sizer = SizerJMX(self.args, self.cluster)
        assert isinstance(sizer, SizerJMX)
        mock_jpype_module.startJVM.assert_called_with("/export/apps/jdk/JDK-1_8_0_72/jre/lib/amd64/server/libjvm.so")
        mock_jpype_module.java.util.HashMap.assert_called_once()

    def test_sizer_missing_username(self):
        self.args.property.append("jmxpass=somepass")
        self.assertRaises(ConfigurationException, SizerJMX, self.args, self.cluster, self.mock_jpype)

    def test_sizer_missing_password(self):
        self.args.property.append("jmxuser=someuser")
        self.assertRaises(ConfigurationException, SizerJMX, self.args, self.cluster, self.mock_jpype)

    def test_sizer_missing_truststore(self):
        self.args.property.append("truststorepass=sometrustpass")
        self.assertRaises(ConfigurationException, SizerJMX, self.args, self.cluster, self.mock_jpype)

    def test_sizer_missing_truststorepass(self):
        self.args.property.append("truststore=somefile")
        self.assertRaises(ConfigurationException, SizerJMX, self.args, self.cluster, self.mock_jpype)

    def test_sizer_set_username_password(self):
        self.args.property.append("jmxuser=someuser")
        self.args.property.append("jmxpass=somepass")
        sizer = SizerJMX(self.args, self.cluster, self.mock_jpype)
        assert sizer._envhash["CREDENTIALS"] == ["someuser", "somepass"]

    def test_sizer_set_truststore(self):
        self.args.property.append("truststore=somefile")
        self.args.property.append("truststorepass=sometrustpass")
        SizerJMX(self.args, self.cluster, self.mock_jpype)
        self.mock_jpype.java.lang.System.setProperty.assert_has_calls(
            [call("javax.net.ssl.trustStore", "somefile"), call("javax.net.ssl.trustStorePassword", "sometrustpass")],
            any_order=True)

    def test_sizer_run_missing_host(self):
        self.cluster.brokers[1].hostname = None
        self.assertRaises(UnknownBrokerException, _validate_broker, self.cluster.brokers[1])

    def test_sizer_run_missing_jmx_port(self):
        self.cluster.brokers[1].jmx_port = -1
        self.assertRaises(UnknownBrokerException, _validate_broker, self.cluster.brokers[1])

    def test_sizer_fetch_bean(self):
        sizer = SizerJMX(self.args, self.cluster, self.mock_jpype)

        bean = MockObjectName("testTopic1", "0")
        connection = MockMBeanServerConnection()
        sizer._fetch_bean(connection, bean)
        assert self.cluster.topics['testTopic1'].partitions[0].size == 1234

    def test_sizer_get_partition_sizes(self):
        sizer = SizerJMX(self.args, self.cluster, self.mock_jpype)

        sizer.get_partition_sizes()
        self.mock_jpype.javax.management.remote.JMXConnectorFactory.connect.assert_called_with("testURL", self.mock_hashmap)
        self.mock_jpype.javax.management.remote.JMXServiceURL.assert_called_with("service:jmx:rmi:///jndi/rmi://brokerhost1.example.com:1099/jmxrmi")
        assert self.cluster.topics['testTopic1'].partitions[0].size == 1234



import unittest

from kafka.tools.protocol.types.integers import Int8


class Int8Tests(unittest.TestCase):
    def test_create(self):
        val = Int8(123)
        assert val.value() == 123

    def test_create_null(self):
        self.assertRaises(TypeError, Int8, None)

    def test_encode_bad_type(self):
        self.assertRaises(TypeError, Int8, 'abc')

    def test_encode_bad_range(self):
        self.assertRaises(ValueError, Int8, 1023)

    def test_decode(self):
        (val, rest) = Int8.decode(b'{')
        assert isinstance(val, Int8)
        assert val.value() == 123
        assert rest == b''

    def test_decode_remainder(self):
        (val, rest) = Int8.decode(b'{abc')
        assert val.value() == 123
        assert rest == b'abc'

    def test_decode_underflow(self):
        self.assertRaises(ValueError, Int8.decode, b'')

    def test_encode(self):
        val = Int8(123)
        assert val.encode() == b'{'



import unittest

from argparse import Namespace
from mock import call, patch, ANY
from subprocess import PIPE
from testfixtures import compare
from testfixtures.popen import MockPopen

from kafka.tools.exceptions import UnknownBrokerException
from kafka.tools.models.broker import Broker
from kafka.tools.models.cluster import Cluster
from kafka.tools.models.topic import Topic
from kafka.tools.assigner.sizers.ssh import SizerSSH


class SizerSSHTests(unittest.TestCase):
    def setUp(self):
        self.args = Namespace()
        self.args.property = ['datadir=/path/to/data']
        self.cluster = self.create_cluster_onehost()

    def create_cluster_onehost(self):
        cluster = Cluster()
        cluster.add_broker(Broker("brokerhost1.example.com", id=1))
        cluster.add_topic(Topic("testTopic1", 2))
        cluster.add_topic(Topic("testTopic2", 2))
        partition = cluster.topics['testTopic1'].partitions[0]
        partition.add_replica(cluster.brokers[1], 0)
        partition = cluster.topics['testTopic1'].partitions[1]
        partition.add_replica(cluster.brokers[1], 0)
        partition = cluster.topics['testTopic2'].partitions[0]
        partition.add_replica(cluster.brokers[1], 0)
        partition = cluster.topics['testTopic2'].partitions[1]
        partition.add_replica(cluster.brokers[1], 0)
        return cluster

    def test_sizer_create(self):
        sizer = SizerSSH(self.args, self.cluster)
        assert isinstance(sizer, SizerSSH)

    def test_sizer_run_missing_host(self):
        self.cluster.brokers[1].hostname = None
        sizer = SizerSSH(self.args, self.cluster)
        self.assertRaises(UnknownBrokerException, sizer.get_partition_sizes)

    def test_sizer_regex_bad(self):
        sizer = SizerSSH(self.args, self.cluster)
        assert sizer.size_re.match("foo") is None

    def test_sizer_regex_good(self):
        sizer = SizerSSH(self.args, self.cluster)
        match_obj = sizer.size_re.match("1001\t/path/to/data/testTopic1-0\n")
        assert match_obj is not None
        assert match_obj.group(1) == "1001"
        assert match_obj.group(2) == "testTopic1"
        assert match_obj.group(3) == "0"

    @patch('kafka.tools.assigner.models.reassignment.subprocess.Popen', new_callable=MockPopen)
    def test_sizer_run_setsizes_singlehost(self, mock_ssh):
        m_stdout = ("1001\t/path/to/data/testTopic1-0\n"
                    "1002\t/path/to/data/testTopic1-1\n"
                    "2001\t/path/to/data/testTopic2-0\n"
                    "2002\t/path/to/data/testTopic2-1\n")
        mock_ssh.set_default(stdout=m_stdout.encode('utf-8'))

        sizer = SizerSSH(self.args, self.cluster)
        sizer.get_partition_sizes()

        compare([call.Popen(['ssh', 'brokerhost1.example.com', 'du -sk /path/to/data/*'], stdout=PIPE, stderr=ANY)], mock_ssh.mock.method_calls)
        assert self.cluster.topics['testTopic1'].partitions[0].size == 1001
        assert self.cluster.topics['testTopic1'].partitions[1].size == 1002
        assert self.cluster.topics['testTopic2'].partitions[0].size == 2001
        assert self.cluster.topics['testTopic2'].partitions[1].size == 2002

    def test_process_df_match(self):
        sizer = SizerSSH(self.args, self.cluster)
        match_obj = sizer.size_re.match("5002\t/path/to/data/testTopic1-1\n")
        sizer.process_df_match(match_obj, 1)
        assert self.cluster.topics['testTopic1'].partitions[1].size == 5002

    def test_process_df_match_badtopic(self):
        sizer = SizerSSH(self.args, self.cluster)
        match_obj = sizer.size_re.match("5002\t/path/to/data/badTopic1-1\n")
        sizer.process_df_match(match_obj, 1)

    def test_process_df_match_badpartition(self):
        sizer = SizerSSH(self.args, self.cluster)
        match_obj = sizer.size_re.match("5002\t/path/to/data/testTopic1-2\n")
        sizer.process_df_match(match_obj, 1)



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.offset_fetch_v0 import _parse_topic_set, OffsetFetchV0Request


class OffsetFetchV0RequestTests(unittest.TestCase):
    def test_parse_topic_set(self):
        val = _parse_topic_set('topicname,4')
        assert val == {'topic': 'topicname', 'partitions': [4]}

    def test_parse_topic_set_multiple(self):
        val = _parse_topic_set('topicname,4,2,8')
        assert val == {'topic': 'topicname', 'partitions': [4, 2, 8]}

    def test_parse_topic_set_nopartitions(self):
        self.assertRaises(ArgumentError, _parse_topic_set, 'topicname')

    def test_parse_topic_set_nonnumeric(self):
        self.assertRaises(ArgumentError, _parse_topic_set, 'topicname,foo,2,8')
        self.assertRaises(ArgumentError, _parse_topic_set, 'topicname,4,foo')

    def test_process_arguments(self):
        val = OffsetFetchV0Request.process_arguments(['groupname', 'topicname,4', 'nexttopic,9'])
        assert val == {'group_id': 'groupname',
                       'topics': [{'topic': 'topicname', 'partitions': [4]},
                                  {'topic': 'nexttopic', 'partitions': [9]}]}

    def test_process_arguments_notenough(self):
        self.assertRaises(ArgumentError, OffsetFetchV0Request.process_arguments, ['groupname'])



import unittest

from argparse import Namespace
from .fixtures import set_up_cluster

from kafka.tools.assigner.actions import ActionModule, ActionBalanceModule


class ActionModuleTests(unittest.TestCase):
    def setUp(self):
        self.cluster = set_up_cluster()
        self.args = Namespace(exclude_topics=[])

    def test_create_class(self):
        action = ActionModule(self.args, self.cluster)
        assert isinstance(action, ActionModule)

    def test_configure_args(self):
        self.assertRaises(Exception, ActionModule.configure_args, None)

    def test_add_args(self):
        action = ActionModule(self.args, self.cluster)
        action._add_args(None)
        assert True

    def test_process_cluster(self):
        action = ActionModule(self.args, self.cluster)
        action.process_cluster()
        assert True

    def test_create_balance_class(self):
        ActionBalanceModule.configure_args(None)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.models import BaseModel
from kafka.tools.protocol.responses.member_assignment_v0 import MemberAssignmentV0


class Group(BaseModel):
    equality_attrs = ['name']

    def __init__(self, name):
        self.name = name
        self.cluster = None
        self.coordinator = None
        self.protocol = None
        self.protocol_type = None
        self.state = None
        self.members = []
        self._last_updated = 0.0

    def updated_since(self, check_time):
        return check_time <= self._last_updated

    def clear_members(self):
        self.members = []

    def add_member(self, name, client_id=None, client_host=None, metadata=None, assignment=None):
        new_member = GroupMember(name, client_id, client_host, metadata, assignment)
        new_member.group = self
        if (self.protocol_type == 'consumer') and (assignment is not None):
            new_member.set_assignment()
        self.members.append(new_member)

    def subscribed_topics(self):
        topics = set()
        for member in self.members:
            for topic in member.topics:
                topics.add(topic)
        return list(topics)


class GroupMember(BaseModel):
    equality_attrs = ['name']

    def __init__(self, name, client_id=None, client_host=None, metadata=None, assignment=None):
        self.name = name
        self.client_id = client_id
        self.client_host = client_host
        self.metadata = metadata
        self.assignment_data = assignment

        self.group = None
        self.user_data = None
        self.assignment_version = None
        self.topics = {}

    def set_assignment(self):
        assignment = MemberAssignmentV0.from_bytes(0, self.assignment_data)
        self.assignment_version = assignment['version'].value()
        self.user_data = assignment['user_data'].value()
        for tp in assignment['partitions']:
            if tp['topic'].value() not in self.topics:
                self.topics[tp['topic'].value()] = []
            for partition in tp['partitions']:
                self.topics[tp['topic'].value()].append(partition.value())



import unittest
from mock import patch

from kafka.tools.protocol.requests.update_metadata_v0 import UpdateMetadataV0Request


class UpdateMetadataV0Tests(unittest.TestCase):
    @patch('kafka.tools.protocol.requests.update_metadata_v0._process_arguments')
    def test_process_arguments_class(self, mock_process):
        mock_process.return_value = 'fake_return_value'
        val = UpdateMetadataV0Request.process_arguments(['fake_arguments'])
        mock_process.assert_called_once_with("UpdateMetadataV0", ['fake_arguments'])
        assert val == 'fake_return_value'



import random
import unittest

from kafka.tools.protocol.types.integers import Int8
from kafka.tools.protocol.types.sequences import Array


class ArrayTests(unittest.TestCase):
    def test_create(self):
        val = Array([1], 'int8')
        assert len(val) == 1

    def test_create_null(self):
        val = Array(None, 'int8')
        assert val.value() is None
        assert len(val) == 0

    def test_create_nonarray(self):
        self.assertRaises(ValueError, Array, 1, 'int8')

    def test_create_badtype(self):
        self.assertRaises(TypeError, Array, [1], 'nonexistant_type')

    def test_decode(self):
        (val, rest) = Array.decode(b'\x00\x00\x00\x01{', schema='int8')
        assert isinstance(val, Array)
        assert len(val) == 1
        assert isinstance(val.value()[0], Int8)
        assert val.value()[0].value() == 123
        assert rest == b''

    def test_decode_bad_type(self):
        self.assertRaises(TypeError, Array.decode, b'\x00\x00\x00\x01{', schema='nonexistant_type')

    def test_decode_two(self):
        (val, rest) = Array.decode(b'\x00\x00\x00\x02{[', schema='int8')
        assert isinstance(val, Array)
        assert len(val) == 2
        assert isinstance(val.value()[0], Int8)
        assert val.value()[0].value() == 123
        assert isinstance(val.value()[1], Int8)
        assert val.value()[1].value() == 91
        assert rest == b''

    def test_decode_remainder(self):
        (val, rest) = Array.decode(b'\x00\x00\x00\x01{abc', schema='int8')
        assert rest == b'abc'

    def test_decode_null(self):
        (val, rest) = Array.decode(b'\xff\xff\xff\xff', schema='int8')
        assert isinstance(val, Array)
        assert val.value() is None
        assert rest == b''

    def test_encode_two(self):
        val = Array([123, 91], 'int8')
        assert val.encode() == b'\x00\x00\x00\x02{['

    def test_encode(self):
        val = Array([123], 'int8')
        assert val.encode() == b'\x00\x00\x00\x01{'

    def test_encode_null(self):
        val = Array(None, 'int8')
        assert val.encode() == b'\xff\xff\xff\xff'

    def test_contains(self):
        i1 = Int8(123)
        i2 = Int8(91)
        i3 = Int8(32)
        val = Array([i1, i2], schema='int8', skip_validation=True)
        assert i1 in val
        assert i2 in val
        assert i3 not in val

    def test_contains_null_array(self):
        i1 = Int8(123)
        val = Array(None, schema='int8')
        assert i1 not in val

    def test_setitem_failure(self):
        obj = Array([123, 91], 'int8')
        self.assertRaises(NotImplementedError, obj.__setitem__, 1, 5)

    def test_delitem_failure(self):
        obj = Array([123, 91], 'int8')
        self.assertRaises(NotImplementedError, obj.__delitem__, 1)

    def test_iterator(self):
        values = [random.randrange(100000) for i in range(100)]
        obj = Array(values, schema='int32')
        t_iter = iter(obj)
        c = 0
        for item in t_iter:
            assert item.value() == values[c]
            c += 1
        assert c == 100

    def test_iterator_reverse(self):
        values = [random.randrange(100000) for i in range(100)]
        obj = Array(values, schema='int32')
        t_iter = reversed(obj)
        c = -1
        for item in t_iter:
            assert item.value() == values[c]
            c -= 1
        assert c == -101

    def test_str(self):
        obj = Array([123, 91], 'int8')
        assert str(obj) == "2 items (array)\n    123 (int8)\n    ---\n    91 (int8)"
        assert repr(obj) == "<Array length=2>"



import re
try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen

from kafka.tools import log
from kafka.tools.exceptions import UnknownBrokerException, ConfigurationException
from kafka.tools.assigner.sizers import SizerModule


class SizerPrometheus(SizerModule):
    name = 'prometheus'
    helpstr = 'Get partition sizes by connection to each broker via Prometheus metrics (https://prometheus.io/)'

    def __init__(self, args, cluster):
        super(SizerPrometheus, self).__init__(args, cluster)
        self._prom_line_re = re.compile(
            r'^(?P<name>[a-zA-Z0-9_]+){(?P<labels>[a-zA-Z0-9]+=\".*\")+(?:,+)?(?:\ +)?} (?P<value>[0-9.E]+)$')
        self._prom_label_re = re.compile(
            r'^(?P<name>[a-zA-Z0-9]+)=\"(?P<value>.*)\"$')

    def _validate_properties(self):
        for propname in ['size_metric_name', 'metrics_port']:
            if propname not in self.properties:
                raise ConfigurationException(
                    "Prometheus sizer requires '{}' property to be set".format(
                        propname))

    def _parse_prometheus_labels(self, label_parts):
        """
        Takes list of labels in Prometheus format and turns those into a dict.
        Example input:
        ['partition="3"', 'topic="mytopic"']
        Will return:
        {
            'partition: '3',
            'topic': 'mytopic'
        }

        :param label_parts: list of label strings
        :return: dict with parsed labels
        """
        labels = {}
        for label in label_parts:
            # be sure to strip spaces around the label
            lm = self._prom_label_re.match(label.strip())
            if lm and len(lm.groups()) == 2:
                labels[lm.group('name')] = lm.group('value')
        return labels

    def _parse_prometheus_value(self, value):
        try:
            val = float(value)
        except ValueError:
            raise UnknownBrokerException(
                "Prometheus sizer cannot convert '{}' to float".format(value))
        return int(val)

    def _parse_prometheus_metric(self, text):
        """
        Parse a metric string in Prometheus format, example:

        kafka_log_size{partition="1",topic="mytopic",} 2.759052553E9

        Return dict with name, labels and value, example:

        {
            'name': 'kafka_log_size',
            'labels': {'partition': '1', 'topic': 'mytopic'},
            'value': 2759052553
        }

        If name, labels or value is missing None will be returned.
        """
        metric = {}
        m = self._prom_line_re.match(text)
        if m and len(m.groups()) == 3:
            metric['name'] = m.group('name')
            metric['value'] = self._parse_prometheus_value(m.group('value'))
            metric['labels'] = self._parse_prometheus_labels(m.group('labels').split(','))

        return metric or None

    def _get_prometheus_metrics(self, hostname, port, path):
        url = 'http://{}:{}{}'.format(hostname, port, path)
        body = ''
        try:
            response = urlopen(url)
            if response.getcode() == 200:
                body = response.read()
            else:
                raise UnknownBrokerException(
                    "Prometheus sizer received invalid (!=200) response code from {}: {}".format(url, response.getcode()))
        except Exception as e:
            raise UnknownBrokerException("Prometheus sizer failed to collect metrics from {}. "
                                         "Error: {}".format(url, e))
        return body.splitlines()

    def _query_prometheus(self, hostname):
        size_metric_name = self.properties['size_metric_name']
        metrics_port = self.properties['metrics_port']
        metrics_path = self.properties.get('metrics_path', '/metrics')
        topic_label = self.properties.get('topic_label', 'topic')
        partition_label = self.properties.get('partition_label', 'partition')
        for text in self._get_prometheus_metrics(hostname, metrics_port, metrics_path):
            m = self._parse_prometheus_metric(text)
            if m and m['name'] == size_metric_name:
                topic = m['labels'][topic_label]
                try:
                    partition = int(m['labels'][partition_label])
                except ValueError:
                    raise UnknownBrokerException(
                        "Prometheus sizer failed to parse partition '{}' as int".format(m['labels'][partition_label]))
                self.cluster.topics[topic].partitions[partition].set_size(m['value'])

    def get_partition_sizes(self):
        self._validate_properties()
        for broker_id, broker in self.cluster.brokers.items():
            if broker.hostname is None:
                raise UnknownBrokerException("Cannot get sizes for broker ID {0} which has no hostname. "
                                             "Remove the broker from the cluster before balance".format(broker_id))

            log.info("Getting partition sizes via Prometheus exporter for {0}".format(broker.hostname))
            self._query_prometheus(broker.hostname)



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.offset_commit_v1 import _get_partition_map, _parse_next_topic, OffsetCommitV1Request


class OffsetCommitV1RequestTests(unittest.TestCase):
    def test_get_partition_map(self):
        val = _get_partition_map([4, 2, 8])
        assert val == {'partition': 4, 'offset': 2, 'timestamp': 8, 'metadata': None}

    def test_get_partition_map_metadata(self):
        val = _get_partition_map([4, 2, 8, 'somemetadata'])
        assert val == {'partition': 4, 'offset': 2, 'timestamp': 8, 'metadata': 'somemetadata'}

    def test_get_partition_map_short(self):
        self.assertRaises(ArgumentError, _get_partition_map, ['foo', 2])

    def test_get_partition_map_extra(self):
        self.assertRaises(ArgumentError, _get_partition_map, ['foo', 2, 8, 'somemetadata', 'foo'])

    def test_get_partition_map_nonnumeric(self):
        self.assertRaises(ArgumentError, _get_partition_map, ['foo', 2, 8])
        self.assertRaises(ArgumentError, _get_partition_map, [4, 'foo', 8])
        self.assertRaises(ArgumentError, _get_partition_map, [4, 2, 'foo'])

    def test_parse_next_topic(self):
        val, rest = _parse_next_topic(['topicname', '4,2,8'])
        assert val == {'topic': 'topicname', 'partitions': [{'partition': 4, 'offset': 2, 'timestamp': 8, 'metadata': None}]}
        assert rest == []

    def test_parse_next_topic_multiple(self):
        val, rest = _parse_next_topic(['topicname', '4,2,8', '9,3,4'])
        assert val == {'topic': 'topicname', 'partitions': [{'partition': 4, 'offset': 2, 'timestamp': 8, 'metadata': None},
                       {'partition': 9, 'offset': 3, 'timestamp': 4, 'metadata': None}]}
        assert rest == []

    def test_parse_next_topic_remainder(self):
        val, rest = _parse_next_topic(['topicname', '4,2,8', 'nexttopic', '9,3,4'])
        assert val == {'topic': 'topicname', 'partitions': [{'partition': 4, 'offset': 2, 'timestamp': 8, 'metadata': None}]}
        assert rest == ['nexttopic', '9,3,4']

    def test_parse_next_topic_nopartitions(self):
        self.assertRaises(ArgumentError, _parse_next_topic, ['topicname'])

    def test_process_arguments(self):
        val = OffsetCommitV1Request.process_arguments(['groupname', '16', 'membername', 'topicname', '4,2,8', 'nexttopic', '9,3,4'])
        assert val == {'group_id': 'groupname',
                       'group_generation_id': 16,
                       'member_id': 'membername',
                       'topics': [{'topic': 'topicname', 'partitions': [{'partition': 4, 'offset': 2, 'timestamp': 8, 'metadata': None}]},
                                  {'topic': 'nexttopic', 'partitions': [{'partition': 9, 'offset': 3, 'timestamp': 4, 'metadata': None}]}]}

    def test_process_arguments_notenough(self):
        self.assertRaises(ArgumentError, OffsetCommitV1Request.process_arguments, ['groupname', '16', 'membername', 'topicname'])

    def test_process_arguments_nonnumeric(self):
        self.assertRaises(ArgumentError, OffsetCommitV1Request.process_arguments, ['groupname', 'foo', 'membername', 'topicname', '4,2,8'])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools import log
from kafka.tools.assigner.actions import ActionModule
from kafka.tools.exceptions import ConfigurationException


class ActionClone(ActionModule):
    name = "clone"
    helpstr = "Copy partitions from some brokers to a new broker (increasing RF)"

    def __init__(self, args, cluster):
        super(ActionClone, self).__init__(args, cluster)

        self.check_brokers()
        if args.to_broker not in self.cluster.brokers:
            raise ConfigurationException("Target broker is not in the brokers list for this cluster")

        self.sources = args.brokers
        self.to_broker = self.cluster.brokers[args.to_broker]

    @classmethod
    def _add_args(cls, parser):
        parser.add_argument('-b', '--brokers', help="List of source broker IDs", required=True, type=int, nargs='*')
        parser.add_argument('-t', '--to_broker', help="Broker ID to copy partitions to", required=True, type=int)

    def process_cluster(self):
        source_set = set(self.sources)
        for partition in self.cluster.partitions(self.args.exclude_topics):
            if len(source_set & set([replica.id for replica in partition.replicas])) > 0:
                if self.to_broker in partition.replicas:
                    log.warn("Target broker (ID {0}) is already in the replica list for {1}:{2}".format(self.to_broker.id, partition.topic.name, partition.num))

                    # If the broker is already in the replica list, it ALWAYS becomes the leader
                    if self.to_broker != partition.replicas[0]:
                        partition.swap_replica_positions(self.to_broker, partition.replicas[0])
                else:
                    # If one of the source brokers is currently the leader, the target broker is the leader. Otherwise, the target leader is in second place
                    if partition.replicas[0].id in self.sources:
                        partition.add_replica(self.to_broker, 0)
                    else:
                        partition.add_replica(self.to_broker, 1)



import sys
import unittest

from mock import patch
from argparse import Namespace
from ..fixtures import set_up_cluster, set_up_subparser

from kafka.tools.assigner.actions.balance import ActionBalance
from kafka.tools.assigner.actions.reorder import ActionReorder
from kafka.tools.assigner.actions.balancemodules.leader import ActionBalanceLeader


class ActionBalanceLeaderTests(unittest.TestCase):
    def setUp(self):
        self.cluster = set_up_cluster()
        (self.parser, self.subparsers) = set_up_subparser()
        self.args = Namespace(exclude_topics=[])

    def test_configure_args(self):
        ActionBalance.configure_args(self.subparsers)
        sys.argv = ['kafka-assigner', 'balance', '-t', 'leader']
        parsed_args = self.parser.parse_args()
        assert parsed_args.action == 'balance'

    def test_create_class(self):
        action = ActionBalanceLeader(self.args, self.cluster)
        assert isinstance(action, ActionBalanceLeader)
        assert isinstance(action._reorder, ActionReorder)

    @patch.object(ActionReorder, 'process_cluster')
    def test_process_cluster(self, mock_reorder):
        self.args.types = ['leader']
        action = ActionBalanceLeader(self.args, self.cluster)
        action.process_cluster()
        mock_reorder.assert_called_once_with()



import argparse

from kafka.tools.models.broker import Broker
from kafka.tools.models.cluster import Cluster
from kafka.tools.models.topic import Topic


def set_up_cluster():
    cluster = Cluster()
    cluster.retention = 100000
    cluster.add_broker(Broker("brokerhost1.example.com", id=1))
    cluster.add_broker(Broker("brokerhost2.example.com", id=2))
    cluster.brokers[1].rack = "a"
    cluster.brokers[2].rack = "b"
    cluster.add_topic(Topic("testTopic1", 2))
    cluster.add_topic(Topic("testTopic2", 2))
    partition = cluster.topics['testTopic1'].partitions[0]
    partition.add_replica(cluster.brokers[1], 0)
    partition.add_replica(cluster.brokers[2], 1)
    partition = cluster.topics['testTopic1'].partitions[1]
    partition.add_replica(cluster.brokers[2], 0)
    partition.add_replica(cluster.brokers[1], 1)
    partition = cluster.topics['testTopic2'].partitions[0]
    partition.add_replica(cluster.brokers[2], 0)
    partition.add_replica(cluster.brokers[1], 1)
    partition = cluster.topics['testTopic2'].partitions[1]
    partition.add_replica(cluster.brokers[1], 0)
    partition.add_replica(cluster.brokers[2], 1)
    return cluster


def set_up_cluster_4broker():
    cluster = Cluster()
    cluster.add_broker(Broker("brokerhost1.example.com", id=1))
    cluster.add_broker(Broker("brokerhost2.example.com", id=2))
    cluster.add_broker(Broker("brokerhost3.example.com", id=3))
    cluster.add_broker(Broker("brokerhost4.example.com", id=4))
    cluster.brokers[1].rack = "a"
    cluster.brokers[2].rack = "a"
    cluster.brokers[3].rack = "b"
    cluster.brokers[4].rack = "b"
    cluster.add_topic(Topic("testTopic1", 4))
    cluster.add_topic(Topic("testTopic2", 4))
    cluster.add_topic(Topic("testTopic3", 4))
    partition = cluster.topics['testTopic1'].partitions[0]
    partition.add_replica(cluster.brokers[1], 0)
    partition.add_replica(cluster.brokers[2], 1)
    partition = cluster.topics['testTopic1'].partitions[1]
    partition.add_replica(cluster.brokers[2], 0)
    partition.add_replica(cluster.brokers[3], 1)
    partition = cluster.topics['testTopic1'].partitions[2]
    partition.add_replica(cluster.brokers[2], 0)
    partition.add_replica(cluster.brokers[3], 1)
    partition = cluster.topics['testTopic1'].partitions[3]
    partition.add_replica(cluster.brokers[4], 0)
    partition.add_replica(cluster.brokers[1], 1)
    partition = cluster.topics['testTopic2'].partitions[0]
    partition.add_replica(cluster.brokers[4], 0)
    partition.add_replica(cluster.brokers[3], 1)
    partition = cluster.topics['testTopic2'].partitions[1]
    partition.add_replica(cluster.brokers[2], 0)
    partition.add_replica(cluster.brokers[4], 1)
    partition = cluster.topics['testTopic2'].partitions[2]
    partition.add_replica(cluster.brokers[2], 0)
    partition.add_replica(cluster.brokers[1], 1)
    partition = cluster.topics['testTopic2'].partitions[3]
    partition.add_replica(cluster.brokers[3], 0)
    partition.add_replica(cluster.brokers[1], 1)
    partition = cluster.topics['testTopic3'].partitions[0]
    partition.add_replica(cluster.brokers[3], 0)
    partition.add_replica(cluster.brokers[2], 1)
    partition = cluster.topics['testTopic3'].partitions[1]
    partition.add_replica(cluster.brokers[4], 0)
    partition.add_replica(cluster.brokers[2], 1)
    partition = cluster.topics['testTopic3'].partitions[2]
    partition.add_replica(cluster.brokers[1], 0)
    partition.add_replica(cluster.brokers[2], 1)
    partition = cluster.topics['testTopic3'].partitions[3]
    partition.add_replica(cluster.brokers[3], 0)
    partition.add_replica(cluster.brokers[4], 1)
    return cluster


def set_up_subparser():
    aparser = argparse.ArgumentParser(prog='kafka-assigner', description='Rejigger Kafka cluster partitions')
    subparsers = aparser.add_subparsers(help='Select manipulation module to use')
    return (aparser, subparsers)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import abc
import six
import struct
from kafka.tools.protocol.types import BaseType


def _get_other_value(other):
    if isinstance(other, BaseIntegerType):
        return other._value
    elif isinstance(other, six.integer_types):
        return other
    else:
        raise TypeError("Cannot compare Integer type to {0}".format(type(other)))


class BaseIntegerType(BaseType):
    @abc.abstractproperty
    def _minimum_value(self):  # pragma: no cover
        raise NotImplementedError

    @abc.abstractproperty
    def _maximum_value(self):  # pragma: no cover
        raise NotImplementedError

    @abc.abstractproperty
    def _num_bytes(self):  # pragma: no cover
        raise NotImplementedError

    @abc.abstractproperty
    def _format_character(self):  # pragma: no cover
        raise NotImplementedError

    def _validate(self):
        if not isinstance(self._value, six.integer_types):
            raise TypeError('Expected "{0}" to be an integer, got: {1}'.format(type(self._value), repr(self._value)))
        if (self._value < self._minimum_value) or (self._value > self._maximum_value):
            raise ValueError("{0} is out of range ({1} ... {2})".format(self._value, self._minimum_value, self._maximum_value))
        return True

    @classmethod
    def decode(cls, byte_array, schema=None):
        if len(byte_array) < cls._num_bytes:
            raise ValueError('Expected {0} byte, only got {1}'.format(cls._num_bytes, len(byte_array)))
        items = struct.unpack(">{0}".format(cls._format_character), byte_array[0:cls._num_bytes])
        return cls(items[0], schema=cls._type), byte_array[cls._num_bytes:]

    def encode(self):
        return struct.pack(">{0}".format(self._format_character), self._value)

    def __str__(self):
        return "{0} ({1})".format(self._value, self._type)

    def __repr__(self):
        return "<{0} {1}>".format(self.__class__.__name__, self._value)

    def __eq__(self, other):
        return self._value == _get_other_value(other)

    def __ne__(self, other):
        return self._value != _get_other_value(other)

    def __lt__(self, other):
        return self._value < _get_other_value(other)

    def __le__(self, other):
        return self._value <= _get_other_value(other)

    def __gt__(self, other):
        return self._value > _get_other_value(other)

    def __ge__(self, other):
        return self._value >= _get_other_value(other)

    def __bool__(self):
        return self._value != 0

    def __nonzero__(self):
        return self.__bool__()


class Int8(BaseIntegerType):
    _type = "int8"
    _minimum_value = -128
    _maximum_value = 127
    _num_bytes = 1
    _format_character = 'b'


class Int16(BaseIntegerType):
    _type = "int16"
    _minimum_value = -32768
    _maximum_value = 32767
    _num_bytes = 2
    _format_character = 'h'


class Int32(BaseIntegerType):
    _type = "int32"
    _minimum_value = -2147483648
    _maximum_value = 2147483647
    _num_bytes = 4
    _format_character = 'i'


class Int64(BaseIntegerType):
    _type = "int64"
    _minimum_value = -9223372036854775808
    _maximum_value = 9223372036854775807
    _num_bytes = 8
    _format_character = 'q'


def decode_length(byte_array, length_cls=Int32):
    val_len, val_data = length_cls.decode(byte_array)
    if val_len.value() > len(val_data):
        raise ValueError('Expected {0} bytes, only got {1}'.format(val_len.value() + length_cls._num_bytes, len(byte_array)))
    return val_len, val_data



import sys
import unittest

from argparse import Namespace
from ..fixtures import set_up_cluster, set_up_subparser

from kafka.tools.models.broker import Broker
from kafka.tools.models.topic import Topic
from kafka.tools.assigner.actions.balance import ActionBalance
from kafka.tools.assigner.actions.balancemodules.rate import ActionBalanceRate


class ActionBalanceRateTests(unittest.TestCase):
    def setUp(self):
        self.cluster = set_up_cluster()
        self.cluster.topics['testTopic1'].partitions[0].set_size(1000)
        self.cluster.topics['testTopic1'].partitions[1].set_size(1000)
        self.cluster.topics['testTopic2'].partitions[0].set_size(2000)
        self.cluster.topics['testTopic2'].partitions[1].set_size(2000)

        (self.parser, self.subparsers) = set_up_subparser()
        self.args = Namespace(exclude_topics=[])

    def test_configure_args(self):
        ActionBalance.configure_args(self.subparsers)
        sys.argv = ['kafka-assigner', 'balance', '-t', 'rate']
        parsed_args = self.parser.parse_args()
        assert parsed_args.action == 'balance'

    def test_create_class(self):
        action = ActionBalanceRate(self.args, self.cluster)
        assert isinstance(action, ActionBalanceRate)

    # This is a duplicate of the size test, but we need at least one test to make sure it does work
    def test_process_cluster_one_move(self):
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        self.cluster.topics['testTopic1'].partitions[0].swap_replica_positions(b1, b2)

        action = ActionBalanceRate(self.args, self.cluster)
        action.process_cluster()

        assert sum([p.scaled_size for p in self.cluster.brokers[1].partitions[0]], 0) == 3000
        assert sum([p.scaled_size for p in self.cluster.brokers[1].partitions[1]], 0) == 3000
        assert sum([p.scaled_size for p in self.cluster.brokers[2].partitions[0]], 0) == 3000
        assert sum([p.scaled_size for p in self.cluster.brokers[2].partitions[1]], 0) == 3000

    def test_process_cluster_zero_partitions(self):
        self.cluster.add_broker(Broker("brokerhost3.example.com", id=3))
        self.cluster.brokers[1].rack = "c"
        self.cluster.add_topic(Topic("testTopic3", 1))
        partition = self.cluster.topics['testTopic3'].partitions[0]
        partition.add_replica(self.cluster.brokers[1], 0)
        partition.add_replica(self.cluster.brokers[2], 1)
        partition.add_replica(self.cluster.brokers[3], 2)
        self.cluster.topics['testTopic3'].partitions[0].set_size(2000)

        self.args.exclude_topics = ['testTopic3']
        action = ActionBalanceRate(self.args, self.cluster)
        action.process_cluster()

        assert sum([p.scaled_size for p in self.cluster.brokers[1].partitions[0]], 0) == 3000
        assert sum([p.scaled_size for p in self.cluster.brokers[1].partitions[1]], 0) == 2000
        assert sum([p.scaled_size for p in self.cluster.brokers[2].partitions[0]], 0) == 3000
        assert sum([p.scaled_size for p in self.cluster.brokers[2].partitions[1]], 0) == 4000
        assert sum([p.scaled_size for p in self.cluster.brokers[3].partitions[0]], 0) == 2000
        assert sum([p.scaled_size for p in self.cluster.brokers[3].partitions[1]], 0) == 2000
        assert sum([p.scaled_size for p in self.cluster.brokers[3].partitions[2]], 0) == 2000



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json
import os
import subprocess
from tempfile import NamedTemporaryFile

from kafka.tools.models import BaseModel


class ReplicaElection(BaseModel):
    equality_attrs = ['partitions']

    def __init__(self, partitions, pause_time=300):
        self.partitions = partitions
        self.pause_time = pause_time

    def __repr__(self):
        return json.dumps(self.dict_for_replica_election())

    def dict_for_replica_election(self):
        ple = {'partitions': []}
        for partition in self.partitions:
            ple['partitions'].append(partition.dict_for_replica_election())
        return ple

    def execute(self, num, total, zookeeper, tools_path, plugins=[], dry_run=True):
        if not dry_run:
            with NamedTemporaryFile(mode='w') as assignfile:
                json.dump(self.dict_for_replica_election(), assignfile)
                assignfile.flush()
                FNULL = open(os.devnull, 'w')
                subprocess.call(['{0}/kafka-preferred-replica-election.sh'.format(tools_path),
                                 '--zookeeper', zookeeper,
                                 '--path-to-json-file', assignfile.name],
                                stdout=FNULL, stderr=FNULL)
